# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation

# Test functions

import pexpect

# default autotest, used to run most tests
# waits for "Test OK"


def default_autotest(child, test_name):
    child.sendline(test_name)
    result = child.expect(["Test OK", "Test Failed",
                           "Command not found", pexpect.TIMEOUT], timeout=900)
    if result == 1:
        return -1, "Fail"
    elif result == 2:
        return -1, "Fail [Not found]"
    elif result == 3:
        return -1, "Fail [Timeout]"
    return 0, "Success"

# autotest used to run dump commands
# just fires the command


def dump_autotest(child, test_name):
    child.sendline(test_name)
    return 0, "Success"

# memory autotest
# reads output and waits for Test OK


def memory_autotest(child, test_name):
    lines = 0
    error = ''
    child.sendline(test_name)
    while True:
        regexp = "IOVA:0x[0-9a-f]*, len:([0-9]*), virt:0x[0-9a-f]*, " \
                 "socket_id:[0-9]*"
        index = child.expect([regexp, "Test OK", "Test Failed",
                              pexpect.TIMEOUT], timeout=10)
        if index == 3:
            return -1, "Fail [Timeout]"
        elif index == 1:
            break
        elif index == 2:
            return -1, "Fail"
        else:
            lines = lines + 1
            size = int(child.match.groups()[0], 10)
            if size <= 0:
                error = 'Bad size'

    if lines <= 0:
        return -1, "Fail [No entries]"
    if error != '':
        return -1, "Fail [{}]".format(error)
    return 0, "Success"


def spinlock_autotest(child, test_name):
    i = 0
    ir = 0
    child.sendline(test_name)
    while True:
        index = child.expect(["Test OK",
                              "Test Failed",
                              "Hello from core ([0-9]*) !",
                              "Hello from within recursive locks "
                              "from ([0-9]*) !",
                              pexpect.TIMEOUT], timeout=5)
        # ok
        if index == 0:
            break

        # message, check ordering
        elif index == 2:
            if int(child.match.groups()[0]) < i:
                return -1, "Fail [Bad order]"
            i = int(child.match.groups()[0])
        elif index == 3:
            if int(child.match.groups()[0]) < ir:
                return -1, "Fail [Bad order]"
            ir = int(child.match.groups()[0])

        # fail
        elif index == 4:
            return -1, "Fail [Timeout]"
        elif index == 1:
            return -1, "Fail"

    return 0, "Success"


def rwlock_autotest(child, test_name):
    i = 0
    child.sendline(test_name)
    while True:
        index = child.expect(["Test OK",
                              "Test Failed",
                              "Hello from core ([0-9]*) !",
                              "Global write lock taken on master "
                              "core ([0-9]*)",
                              pexpect.TIMEOUT], timeout=10)
        # ok
        if index == 0:
            if i != 0xffff:
                return -1, "Fail [Message is missing]"
            break

        # message, check ordering
        elif index == 2:
            if int(child.match.groups()[0]) < i:
                return -1, "Fail [Bad order]"
            i = int(child.match.groups()[0])

        # must be the last message, check ordering
        elif index == 3:
            i = 0xffff

        elif index == 4:
            return -1, "Fail [Timeout]"

        # fail
        else:
            return -1, "Fail"

    return 0, "Success"


def ticketlock_autotest(child, test_name):
    i = 0
    ir = 0
    child.sendline(test_name)
    while True:
        index = child.expect(["Test OK",
                              "Test Failed",
                              "Hello from core ([0-9]*) !",
                              "Hello from within recursive locks "
                              "from ([0-9]*) !",
                              pexpect.TIMEOUT], timeout=5)
        # ok
        if index == 0:
            break

        # message, check ordering
        elif index == 2:
            if int(child.match.groups()[0]) < i:
                return -1, "Fail [Bad order]"
            i = int(child.match.groups()[0])
        elif index == 3:
            if int(child.match.groups()[0]) < ir:
                return -1, "Fail [Bad order]"
            ir = int(child.match.groups()[0])

        # fail
        elif index == 4:
            return -1, "Fail [Timeout]"
        elif index == 1:
            return -1, "Fail"

    return 0, "Success"

def mcslock_autotest(child, test_name):
    i = 0
    ir = 0
    child.sendline(test_name)
    while True:
        index = child.expect(["Test OK",
                              "Test Failed",
                              "lcore ([0-9]*) state: ([0-1])"
                              "MCS lock taken on core ([0-9]*)",
                              "MCS lock released on core ([0-9]*)",
                              pexpect.TIMEOUT], timeout=5)
        # ok
        if index == 0:
            break

        # message, check ordering
        elif index == 2:
            if int(child.match.groups()[0]) < i:
                return -1, "Fail [Bad order]"
            i = int(child.match.groups()[0])
        elif index == 3:
            if int(child.match.groups()[0]) < ir:
                return -1, "Fail [Bad order]"
            ir = int(child.match.groups()[0])

        # fail
        elif index == 4:
            return -1, "Fail [Timeout]"
        elif index == 1:
            return -1, "Fail"

    return 0, "Success"

def logs_autotest(child, test_name):
    child.sendline(test_name)

    log_list = [
        "TESTAPP1: error message",
        "TESTAPP1: critical message",
        "TESTAPP2: critical message",
        "TESTAPP1: error message",
    ]

    for log_msg in log_list:
        index = child.expect([log_msg,
                              "Test OK",
                              "Test Failed",
                              pexpect.TIMEOUT], timeout=10)

        if index == 3:
            return -1, "Fail [Timeout]"
        # not ok
        elif index != 0:
            return -1, "Fail"

    index = child.expect(["Test OK",
                          "Test Failed",
                          pexpect.TIMEOUT], timeout=10)

    return 0, "Success"


def timer_autotest(child, test_name):
    child.sendline(test_name)

    index = child.expect(["Start timer stress tests",
                          "Test Failed",
                          pexpect.TIMEOUT], timeout=5)

    if index == 1:
        return -1, "Fail"
    elif index == 2:
        return -1, "Fail [Timeout]"

    index = child.expect(["Start timer stress tests 2",
                          "Test Failed",
                          pexpect.TIMEOUT], timeout=5)

    if index == 1:
        return -1, "Fail"
    elif index == 2:
        return -1, "Fail [Timeout]"

    index = child.expect(["Start timer basic tests",
                          "Test Failed",
                          pexpect.TIMEOUT], timeout=5)

    if index == 1:
        return -1, "Fail"
    elif index == 2:
        return -1, "Fail [Timeout]"

    lcore_tim0 = -1
    lcore_tim1 = -1
    lcore_tim2 = -1
    lcore_tim3 = -1

    while True:
        index = child.expect(["TESTTIMER: ([0-9]*): callback id=([0-9]*) "
                              "count=([0-9]*) on core ([0-9]*)",
                              "Test OK",
                              "Test Failed",
                              pexpect.TIMEOUT], timeout=10)

        if index == 1:
            break

        if index == 2:
            return -1, "Fail"
        elif index == 3:
            return -1, "Fail [Timeout]"

        try:
            id = int(child.match.groups()[1])
            cnt = int(child.match.groups()[2])
            lcore = int(child.match.groups()[3])
        except:
            return -1, "Fail [Cannot parse]"

        # timer0 always expires on the same core when cnt < 20
        if id == 0:
            if lcore_tim0 == -1:
                lcore_tim0 = lcore
            elif lcore != lcore_tim0 and cnt < 20:
                return -1, "Fail [lcore != lcore_tim0 (%d, %d)]" \
                    % (lcore, lcore_tim0)
            if cnt > 21:
                return -1, "Fail [tim0 cnt > 21]"

        # timer1 each time expires on a different core
        if id == 1:
            if lcore == lcore_tim1:
                return -1, "Fail [lcore == lcore_tim1 (%d, %d)]" \
                    % (lcore, lcore_tim1)
            lcore_tim1 = lcore
            if cnt > 10:
                return -1, "Fail [tim1 cnt > 30]"

        # timer0 always expires on the same core
        if id == 2:
            if lcore_tim2 == -1:
                lcore_tim2 = lcore
            elif lcore != lcore_tim2:
                return -1, "Fail [lcore != lcore_tim2 (%d, %d)]" \
                    % (lcore, lcore_tim2)
            if cnt > 30:
                return -1, "Fail [tim2 cnt > 30]"

        # timer0 always expires on the same core
        if id == 3:
            if lcore_tim3 == -1:
                lcore_tim3 = lcore
            elif lcore != lcore_tim3:
                return -1, "Fail [lcore_tim3 changed (%d -> %d)]" \
                    % (lcore, lcore_tim3)
            if cnt > 30:
                return -1, "Fail [tim3 cnt > 30]"

    # must be 2 different cores
    if lcore_tim0 == lcore_tim3:
        return -1, "Fail [lcore_tim0 (%d) == lcore_tim3 (%d)]" \
            % (lcore_tim0, lcore_tim3)

    return 0, "Success"


def ring_autotest(child, test_name):
    child.sendline(test_name)
    index = child.expect(["Test OK", "Test Failed",
                          pexpect.TIMEOUT], timeout=2)
    if index == 1:
        return -1, "Fail"
    elif index == 2:
        return -1, "Fail [Timeout]"

    return 0, "Success"



"""
Rbd testing task
"""
import contextlib
import logging
import os
import tempfile
import sys

from io import StringIO
from teuthology.orchestra import run
from teuthology import misc as teuthology
from teuthology import contextutil
from teuthology.parallel import parallel
from teuthology.task.common_fs_utils import generic_mkfs
from teuthology.task.common_fs_utils import generic_mount
from teuthology.task.common_fs_utils import default_image_name


#V1 image unsupported but required for testing purposes
os.environ["RBD_FORCE_ALLOW_V1"] = "1"

log = logging.getLogger(__name__)

ENCRYPTION_PASSPHRASE = "password"
CLONE_ENCRYPTION_PASSPHRASE = "password2"

@contextlib.contextmanager
def create_image(ctx, config):
    """
    Create an rbd image.

    For example::

        tasks:
        - ceph:
        - rbd.create_image:
            client.0:
                image_name: testimage
                image_size: 100
                image_format: 1
                encryption_format: luks2
            client.1:

    Image size is expressed as a number of megabytes; default value
    is 10240.

    Image format value must be either 1 or 2; default value is 1.

    """
    assert isinstance(config, dict) or isinstance(config, list), \
        "task create_image only supports a list or dictionary for configuration"

    if isinstance(config, dict):
        images = config.items()
    else:
        images = [(role, None) for role in config]

    testdir = teuthology.get_testdir(ctx)
    passphrase_file = '{tdir}/passphrase'.format(tdir=testdir)
    for role, properties in images:
        if properties is None:
            properties = {}
        name = properties.get('image_name', default_image_name(role))
        size = properties.get('image_size', 10240)
        fmt = properties.get('image_format', 1)
        encryption_format = properties.get('encryption_format', 'none')
        (remote,) = ctx.cluster.only(role).remotes.keys()
        log.info('Creating image {name} with size {size}'.format(name=name,
                                                                 size=size))
        args = [
                'adjust-ulimits',
                'ceph-coverage',
                '{tdir}/archive/coverage'.format(tdir=testdir),
                'rbd',
                '-p', 'rbd',
                'create',
                '--size', str(size),
                name,
            ]
        # omit format option if using the default (format 1)
        # since old versions of don't support it
        if int(fmt) != 1:
            args += ['--image-format', str(fmt)]
        remote.run(args=args)

        if encryption_format != 'none':
            remote.run(
                args=[
                    'echo',
                    ENCRYPTION_PASSPHRASE,
                    run.Raw('>'),
                    passphrase_file
                    ]
                )
            remote.run(
                args=[
                    'adjust-ulimits',
                    'ceph-coverage',
                    '{tdir}/archive/coverage'.format(tdir=testdir),
                    'rbd',
                    'encryption',
                    'format',
                    name,
                    encryption_format,
                    passphrase_file,
                    '-p',
                    'rbd'
                    ]
                )
    try:
        yield
    finally:
        log.info('Deleting rbd images...')
        remote.run(args=['rm', '-f', passphrase_file])
        for role, properties in images:
            if properties is None:
                properties = {}
            name = properties.get('image_name', default_image_name(role))
            (remote,) = ctx.cluster.only(role).remotes.keys()
            remote.run(
                args=[
                    'adjust-ulimits',
                    'ceph-coverage',
                    '{tdir}/archive/coverage'.format(tdir=testdir),
                    'rbd',
                    '-p', 'rbd',
                    'rm',
                    name,
                    ],
                )

@contextlib.contextmanager
def clone_image(ctx, config):
    """
    Clones a parent imag

    For example::

        tasks:
        - ceph:
        - rbd.clone_image:
            client.0:
                parent_name: testimage
                image_name: cloneimage
                encryption_format: luks2
    """
    assert isinstance(config, dict) or isinstance(config, list), \
        "task clone_image only supports a list or dictionary for configuration"

    if isinstance(config, dict):
        images = config.items()
    else:
        images = [(role, None) for role in config]

    testdir = teuthology.get_testdir(ctx)
    clone_passphrase_file = '{tdir}/clone-passphrase'.format(tdir=testdir)
    for role, properties in images:
        if properties is None:
            properties = {}

        name = properties.get('image_name', default_image_name(role))
        parent_name = properties.get('parent_name')
        assert parent_name is not None, \
            "parent_name is required"
        parent_spec = '{name}@{snap}'.format(name=parent_name, snap=name)

        (remote,) = ctx.cluster.only(role).remotes.keys()
        log.info('Clone image {parent} to {child}'.format(parent=parent_name,
                                                          child=name))

        commands = [('snap', 'create', parent_spec),
                    ('snap', 'protect', parent_spec),
                    ('clone', parent_spec, name)
                    ]

        encryption_format = properties.get('encryption_format', 'none')
        if encryption_format != 'none':
            remote.run(
                args=[
                    'echo',
                    CLONE_ENCRYPTION_PASSPHRASE,
                    run.Raw('>'),
                    clone_passphrase_file
                    ]
                )

            commands.append(
                ('encryption', 'format', name, encryption_format,
                 clone_passphrase_file)
            )

        for cmd in commands:
            args = [
                    'adjust-ulimits',
                    'ceph-coverage',
                    '{tdir}/archive/coverage'.format(tdir=testdir),
                    'rbd', '-p', 'rbd'
                    ]
            args.extend(cmd)
            remote.run(args=args)

    try:
        yield
    finally:
        log.info('Deleting rbd clones...')
        remote.run(args=['rm', '-f', clone_passphrase_file])
        for role, properties in images:
            if properties is None:
                properties = {}
            name = properties.get('image_name', default_image_name(role))
            parent_name = properties.get('parent_name')
            parent_spec = '{name}@{snap}'.format(name=parent_name, snap=name)

            (remote,) = ctx.cluster.only(role).remotes.keys()

            for cmd in [('rm', name),
                        ('snap', 'unprotect', parent_spec),
                        ('snap', 'rm', parent_spec)]:
                args = [
                        'adjust-ulimits',
                        'ceph-coverage',
                        '{tdir}/archive/coverage'.format(tdir=testdir),
                        'rbd', '-p', 'rbd'
                        ]
                args.extend(cmd)
                remote.run(args=args)

@contextlib.contextmanager
def modprobe(ctx, config):
    """
    Load the rbd kernel module..

    For example::

        tasks:
        - ceph:
        - rbd.create_image: [client.0]
        - rbd.modprobe: [client.0]
    """
    log.info('Loading rbd kernel module...')
    for role in config:
        (remote,) = ctx.cluster.only(role).remotes.keys()
        remote.run(
            args=[
                'sudo',
                'modprobe',
                'rbd',
                ],
            )
    try:
        yield
    finally:
        log.info('Unloading rbd kernel module...')
        for role in config:
            (remote,) = ctx.cluster.only(role).remotes.keys()
            remote.run(
                args=[
                    'sudo',
                    'modprobe',
                    '-r',
                    'rbd',
                    # force errors to be ignored; necessary if more
                    # than one device was created, which may mean
                    # the module isn't quite ready to go the first
                    # time through.
                    run.Raw('||'),
                    'true',
                    ],
                )

@contextlib.contextmanager
def dev_create(ctx, config):
    """
    Map block devices to rbd images.

    For example::

        tasks:
        - ceph:
        - rbd.create_image: [client.0]
        - rbd.modprobe: [client.0]
        - rbd.dev_create:
            client.0:
                image_name: testimage.client.0
                encryption_format: luks2
                parent_encryption_format: luks1
    """
    assert isinstance(config, dict) or isinstance(config, list), \
        "task dev_create only supports a list or dictionary for configuration"

    if isinstance(config, dict):
        images = config.items()
    else:
        images = [(role, None) for role in config]

    log.info('Creating rbd block devices...')

    testdir = teuthology.get_testdir(ctx)
    passphrase_file = '{tdir}/passphrase'.format(tdir=testdir)
    clone_passphrase_file = '{tdir}/clone-passphrase'.format(tdir=testdir)
    device_path = {}

    for role, properties in images:
        if properties is None:
            properties = {}
        name = properties.get('image_name', default_image_name(role))
        parent_encryption_format = properties.get('parent_encryption_format',
                                                  'none')
        encryption_format = properties.get('encryption_format',
                                           parent_encryption_format)
        (remote,) = ctx.cluster.only(role).remotes.keys()

        if encryption_format == 'none' and parent_encryption_format == 'none':
            device_path[role] = '/dev/rbd/rbd/{image}'.format(image=name)
            device_specific_args = []
        else:
            device_specific_args = ['-t', 'nbd', '-o']

            is_cloned = properties.get('parent_name') is not None
            encryption_args = ""
            if is_cloned and properties.get('encryption_format') != 'none':
                remote.run(
                    args=[
                        'echo',
                        CLONE_ENCRYPTION_PASSPHRASE,
                        run.Raw('>'),
                        clone_passphrase_file
                        ]
                    )

                encryption_args = \
                    'encryption-format=%s,encryption-passphrase-file=%s' % (
                        encryption_format, clone_passphrase_file)

            if not is_cloned or parent_encryption_format != 'none':
                remote.run(
                    args=[
                        'echo',
                        ENCRYPTION_PASSPHRASE,
                        run.Raw('>'),
                        passphrase_file
                        ]
                    )

                if is_cloned and properties.get('encryption_format') != 'none':
                    encryption_args += ","

                if parent_encryption_format != 'none':
                    encryption_args += \
                        'encryption-format=%s,encryption-passphrase-file=%s' % (
                            parent_encryption_format, passphrase_file)
                else:
                    encryption_args += \
                        'encryption-format=%s,encryption-passphrase-file=%s' % (
                            encryption_format, passphrase_file)

            device_specific_args.append(encryption_args)

        map_fp = StringIO()
        remote.run(
            args=[
                'sudo',
                'adjust-ulimits',
                'ceph-coverage',
                '{tdir}/archive/coverage'.format(tdir=testdir),
                'rbd',
                '--id', role.rsplit('.')[-1],
                '-p', 'rbd',
                'map',
                name] + device_specific_args,
            stdout=map_fp,
            )

        if encryption_format != 'none' or parent_encryption_format != 'none':
            device_path[role] = map_fp.getvalue().rstrip()
            properties['device_path'] = device_path[role]
            remote.run(args=['sudo', 'chmod', '666', device_path[role]])
    try:
        yield
    finally:
        log.info('Unmapping rbd devices...')
        remote.run(args=['rm', '-f', passphrase_file, clone_passphrase_file])
        for role, properties in images:
            if not device_path.get(role):
                continue

            if properties is None:
                properties = {}
            encryption_format = properties.get('encryption_format', 'none')
            parent_encryption_format = properties.get(
                'parent_encryption_format', 'none')
            (remote,) = ctx.cluster.only(role).remotes.keys()

            if encryption_format == 'none' and \
                    parent_encryption_format == 'none':
                device_specific_args = []
            else:
                device_specific_args = ['-t', 'nbd']

            remote.run(
                args=[
                    'LD_LIBRARY_PATH={tdir}/binary/usr/local/lib'.format(tdir=testdir),
                    'sudo',
                    'adjust-ulimits',
                    'ceph-coverage',
                    '{tdir}/archive/coverage'.format(tdir=testdir),
                    'rbd',
                    '-p', 'rbd',
                    'unmap',
                    device_path[role],
                    ] + device_specific_args,
                )


def rbd_devname_rtn(ctx, image):
    return '/dev/rbd/rbd/{image}'.format(image=image)    

def canonical_path(ctx, role, path):
    """
    Determine the canonical path for a given path on the host
    representing the given role.  A canonical path contains no
    . or .. components, and includes no symbolic links.
    """
    version_fp = StringIO()
    ctx.cluster.only(role).run(
        args=[ 'readlink', '-f', path ],
        stdout=version_fp,
        )
    canonical_path = version_fp.getvalue().rstrip('\n')
    version_fp.close()
    return canonical_path

@contextlib.contextmanager
def run_xfstests(ctx, config):
    """
    Run xfstests over specified devices.

    Warning: both the test and scratch devices specified will be
    overwritten.  Normally xfstests modifies (but does not destroy)
    the test device, but for now the run script used here re-makes
    both filesystems.

    Note: Only one instance of xfstests can run on a single host at
    a time, although this is not enforced.

    This task in its current form needs some improvement.  For
    example, it assumes all roles provided in the config are
    clients, and that the config provided is a list of key/value
    pairs.  For now please use the xfstests() interface, below.

    For example::

        tasks:
        - ceph:
        - rbd.run_xfstests:
            client.0:
                count: 2
                test_dev: 'test_dev'
                scratch_dev: 'scratch_dev'
                fs_type: 'xfs'
                tests: 'generic/100 xfs/003 xfs/005 xfs/006 generic/015'
                exclude:
                - generic/42
                randomize: true
    """
    with parallel() as p:
        for role, properties in config.items():
            p.spawn(run_xfstests_one_client, ctx, role, properties)
        exc = None
        while True:
            try:
                p.next()
            except StopIteration:
                break
            except:
                exc = sys.exc_info()[1]
        if exc is not None:
            raise exc
    yield

def run_xfstests_one_client(ctx, role, properties):
    """
    Spawned routine to handle xfs tests for a single client
    """
    testdir = teuthology.get_testdir(ctx)
    try:
        count = properties.get('count')
        test_dev = properties.get('test_dev')
        assert test_dev is not None, \
            "task run_xfstests requires test_dev to be defined"
        test_dev = canonical_path(ctx, role, test_dev)

        scratch_dev = properties.get('scratch_dev')
        assert scratch_dev is not None, \
            "task run_xfstests requires scratch_dev to be defined"
        scratch_dev = canonical_path(ctx, role, scratch_dev)

        fs_type = properties.get('fs_type')
        tests = properties.get('tests')
        exclude_list = properties.get('exclude')
        randomize = properties.get('randomize')

        (remote,) = ctx.cluster.only(role).remotes.keys()

        # Fetch the test script
        test_root = teuthology.get_testdir(ctx)
        test_script = 'run_xfstests.sh'
        test_path = os.path.join(test_root, test_script)

        xfstests_url = properties.get('xfstests_url')
        assert xfstests_url is not None, \
            "task run_xfstests requires xfstests_url to be defined"

        xfstests_krbd_url = xfstests_url + '/' + test_script

        log.info('Fetching {script} for {role} from {url}'.format(
            script=test_script,
            role=role,
            url=xfstests_krbd_url))

        args = [ 'wget', '-O', test_path, '--', xfstests_krbd_url ]
        remote.run(args=args)

        log.info('Running xfstests on {role}:'.format(role=role))
        log.info('   iteration count: {count}:'.format(count=count))
        log.info('       test device: {dev}'.format(dev=test_dev))
        log.info('    scratch device: {dev}'.format(dev=scratch_dev))
        log.info('     using fs_type: {fs_type}'.format(fs_type=fs_type))
        log.info('      tests to run: {tests}'.format(tests=tests))
        log.info('      exclude list: {}'.format(' '.join(exclude_list)))
        log.info('         randomize: {randomize}'.format(randomize=randomize))

        if exclude_list:
            with tempfile.NamedTemporaryFile(mode='w', prefix='exclude') as exclude_file:
                for test in exclude_list:
                    exclude_file.write("{}\n".format(test))
                exclude_file.flush()
                remote.put_file(exclude_file.name, exclude_file.name)

        # Note that the device paths are interpreted using
        # readlink -f <path> in order to get their canonical
        # pathname (so it matches what the kernel remembers).
        args = [
            '/usr/bin/sudo',
            'TESTDIR={tdir}'.format(tdir=testdir),
            'adjust-ulimits',
            'ceph-coverage',
            '{tdir}/archive/coverage'.format(tdir=testdir),
            '/bin/bash',
            test_path,
            '-c', str(count),
            '-f', fs_type,
            '-t', test_dev,
            '-s', scratch_dev,
            ]
        if exclude_list:
            args.extend(['-x', exclude_file.name])
        if randomize:
            args.append('-r')
        if tests:
            args.extend(['--', tests])
        remote.run(args=args, logger=log.getChild(role))
    finally:
        log.info('Removing {script} on {role}'.format(script=test_script,
                                                      role=role))
        remote.run(args=['rm', '-f', test_path])

@contextlib.contextmanager
def xfstests(ctx, config):
    """
    Run xfstests over rbd devices.  This interface sets up all
    required configuration automatically if not otherwise specified.
    Note that only one instance of xfstests can run on a single host
    at a time.  By default, the set of tests specified is run once.
    If a (non-zero) count value is supplied, the complete set of
    tests will be run that number of times.

    For example::

        tasks:
        - ceph:
        # Image sizes are in MB
        - rbd.xfstests:
            client.0:
                count: 3
                test_image: 'test_image'
                test_size: 250
                test_format: 2
                scratch_image: 'scratch_image'
                scratch_size: 250
                scratch_format: 1
                fs_type: 'xfs'
                tests: 'generic/100 xfs/003 xfs/005 xfs/006 generic/015'
                exclude:
                - generic/42
                randomize: true
                xfstests_url: 'https://raw.github.com/ceph/ceph-ci/wip-55555/qa'
    """
    if config is None:
        config = { 'all': None }
    assert isinstance(config, dict) or isinstance(config, list), \
        "task xfstests only supports a list or dictionary for configuration"
    if isinstance(config, dict):
        config = teuthology.replace_all_with_clients(ctx.cluster, config)
        runs = config.items()
    else:
        runs = [(role, None) for role in config]

    running_xfstests = {}
    for role, properties in runs:
        assert role.startswith('client.'), \
            "task xfstests can only run on client nodes"
        for host, roles_for_host in ctx.cluster.remotes.items():
            if role in roles_for_host:
                assert host not in running_xfstests, \
                    "task xfstests allows only one instance at a time per host"
                running_xfstests[host] = True

    images_config = {}
    scratch_config = {}
    modprobe_config = {}
    image_map_config = {}
    scratch_map_config = {}
    xfstests_config = {}
    for role, properties in runs:
        if properties is None:
            properties = {}

        test_image = properties.get('test_image', 'test_image.{role}'.format(role=role))
        test_size = properties.get('test_size', 10000) # 10G
        test_fmt = properties.get('test_format', 1)
        scratch_image = properties.get('scratch_image', 'scratch_image.{role}'.format(role=role))
        scratch_size = properties.get('scratch_size', 10000) # 10G
        scratch_fmt = properties.get('scratch_format', 1)

        images_config[role] = dict(
            image_name=test_image,
            image_size=test_size,
            image_format=test_fmt,
            )

        scratch_config[role] = dict(
            image_name=scratch_image,
            image_size=scratch_size,
            image_format=scratch_fmt,
            )

        xfstests_branch = properties.get('xfstests_branch', 'master')
        xfstests_url = properties.get('xfstests_url', 'https://raw.github.com/ceph/ceph/{branch}/qa'.format(branch=xfstests_branch))

        xfstests_config[role] = dict(
            count=properties.get('count', 1),
            test_dev='/dev/rbd/rbd/{image}'.format(image=test_image),
            scratch_dev='/dev/rbd/rbd/{image}'.format(image=scratch_image),
            fs_type=properties.get('fs_type', 'xfs'),
            randomize=properties.get('randomize', False),
            tests=properties.get('tests'),
            exclude=properties.get('exclude', []),
            xfstests_url=xfstests_url,
            )

        log.info('Setting up xfstests using RBD images:')
        log.info('      test ({size} MB): {image}'.format(size=test_size,
                                                        image=test_image))
        log.info('   scratch ({size} MB): {image}'.format(size=scratch_size,
                                                        image=scratch_image))
        modprobe_config[role] = None
        image_map_config[role] = {'image_name': test_image}
        scratch_map_config[role] = {'image_name': scratch_image}

    with contextutil.nested(
        lambda: create_image(ctx=ctx, config=images_config),
        lambda: create_image(ctx=ctx, config=scratch_config),
        lambda: modprobe(ctx=ctx, config=modprobe_config),
        lambda: dev_create(ctx=ctx, config=image_map_config),
        lambda: dev_create(ctx=ctx, config=scratch_map_config),
        lambda: run_xfstests(ctx=ctx, config=xfstests_config),
        ):
        yield


@contextlib.contextmanager
def task(ctx, config):
    """
    Create and mount an rbd image.

    For example, you can specify which clients to run on::

        tasks:
        - ceph:
        - rbd: [client.0, client.1]

    There are a few image options::

        tasks:
        - ceph:
        - rbd:
            client.0: # uses defaults
            client.1:
                image_name: foo
                image_size: 2048
                image_format: 2
                fs_type: xfs

    To use default options on all clients::

        tasks:
        - ceph:
        - rbd:
            all:

    To create 20GiB images and format them with xfs on all clients::

        tasks:
        - ceph:
        - rbd:
            all:
              image_size: 20480
              fs_type: xfs
    """
    if config is None:
        config = { 'all': None }
    norm_config = config
    if isinstance(config, dict):
        norm_config = teuthology.replace_all_with_clients(ctx.cluster, config)
    if isinstance(norm_config, dict):
        role_images = {}
        for role, properties in norm_config.items():
            if properties is None:
                properties = {}
            role_images[role] = properties.get('image_name')
    else:
        role_images = norm_config

    log.debug('rbd config is: %s', norm_config)

    with contextutil.nested(
        lambda: create_image(ctx=ctx, config=norm_config),
        lambda: modprobe(ctx=ctx, config=norm_config),
        lambda: dev_create(ctx=ctx, config=norm_config),
        lambda: generic_mkfs(ctx=ctx, config=norm_config,
                devname_rtn=rbd_devname_rtn),
        lambda: generic_mount(ctx=ctx, config=role_images,
                devname_rtn=rbd_devname_rtn),
        ):
        yield



#!/usr/bin/env python
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation

# Script that runs cmdline_test app and feeds keystrokes into it.
from __future__ import print_function
import cmdline_test_data
import os
import pexpect
import sys


#
# function to run test
#
def runTest(child, test):
    child.send(test["Sequence"])
    if test["Result"] is None:
        return 0
    child.expect(test["Result"], 1)


#
# history test is a special case
#
# This test does the following:
# 1) fills the history with garbage up to its full capacity
#    (just enough to remove last entry)
# 2) scrolls back history to the very beginning
# 3) checks if the output is as expected, that is, the first
#    number in the sequence (not the last entry before it)
#
# This is a self-contained test, it needs only a pexpect child
#
def runHistoryTest(child):
    # find out history size
    child.sendline(cmdline_test_data.CMD_GET_BUFSIZE)
    child.expect("History buffer size: \\d+", timeout=1)
    history_size = int(child.after[len(cmdline_test_data.BUFSIZE_TEMPLATE):])
    i = 0

    # fill the history with numbers
    while i < history_size / 10:
        # add 1 to prevent from parsing as octals
        child.send("1" + str(i).zfill(8) + cmdline_test_data.ENTER)
        # the app will simply print out the number
        child.expect(str(i + 100000000), timeout=1)
        i += 1
    # scroll back history
    child.send(cmdline_test_data.UP * (i + 2) + cmdline_test_data.ENTER)
    child.expect("100000000", timeout=1)

# the path to cmdline_test executable is supplied via command-line.
if len(sys.argv) < 2:
    print("Error: please supply cmdline_test app path")
    sys.exit(1)

test_app_path = sys.argv[1]

if not os.path.exists(test_app_path):
    print("Error: please supply cmdline_test app path")
    sys.exit(1)

child = pexpect.spawn(test_app_path)

print("Running command-line tests...")
for test in cmdline_test_data.tests:
    testname = (test["Name"] + ":").ljust(30)
    try:
        runTest(child, test)
        print(testname, "PASS")
    except:
        print(testname, "FAIL")
        print(child)
        sys.exit(1)

# since last test quits the app, run new instance
child = pexpect.spawn(test_app_path)

testname = ("History fill test:").ljust(30)
try:
    runHistoryTest(child)
    print(testname, "PASS")
except:
    print(testname, "FAIL")
    print(child)
    sys.exit(1)
child.close()
sys.exit(0)



#!/usr/bin/env python
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation
# Copyright(c) 2017 Cavium, Inc. All rights reserved.

from __future__ import print_function
import sys
try:
    xrange # Python 2
except NameError:
    xrange = range # Python 3

sockets = []
cores = []
core_map = {}
base_path = "/sys/devices/system/cpu"
fd = open("{}/kernel_max".format(base_path))
max_cpus = int(fd.read())
fd.close()
for cpu in xrange(max_cpus + 1):
    try:
        fd = open("{}/cpu{}/topology/core_id".format(base_path, cpu))
    except IOError:
        continue
    except:
        break
    core = int(fd.read())
    fd.close()
    fd = open("{}/cpu{}/topology/physical_package_id".format(base_path, cpu))
    socket = int(fd.read())
    fd.close()
    if core not in cores:
        cores.append(core)
    if socket not in sockets:
        sockets.append(socket)
    key = (socket, core)
    if key not in core_map:
        core_map[key] = []
    core_map[key].append(cpu)

print(format("=" * (47 + len(base_path))))
print("Core and Socket Information (as reported by '{}')".format(base_path))
print("{}\n".format("=" * (47 + len(base_path))))
print("cores = ", cores)
print("sockets = ", sockets)
print("")

max_processor_len = len(str(len(cores) * len(sockets) * 2 - 1))
max_thread_count = len(list(core_map.values())[0])
max_core_map_len = (max_processor_len * max_thread_count)  \
                      + len(", ") * (max_thread_count - 1) \
                      + len('[]') + len('Socket ')
max_core_id_len = len(str(max(cores)))

output = " ".ljust(max_core_id_len + len('Core '))
for s in sockets:
    output += " Socket %s" % str(s).ljust(max_core_map_len - len('Socket '))
print(output)

output = " ".ljust(max_core_id_len + len('Core '))
for s in sockets:
    output += " --------".ljust(max_core_map_len)
    output += " "
print(output)

for c in cores:
    output = "Core %s" % str(c).ljust(max_core_id_len)
    for s in sockets:
        if (s,c) in core_map:
            output += " " + str(core_map[(s, c)]).ljust(max_core_map_len)
        else:
            output += " " * (max_core_map_len + 1)
    print(output)



import errno

from io import StringIO
from time import sleep
from logging import getLogger

from tasks.cephfs.cephfs_test_case import CephFSTestCase
from tasks.cephfs.caps_helper import (CapTester, gen_mon_cap_str,
                                      gen_osd_cap_str, gen_mds_cap_str)


log = getLogger(__name__)


class SwapHelper(CephFSTestCase):
    '''
    Helper methods for testing "ceph fs swap" command.
    '''

    MDSS_REQUIRED = 3
    CLIENTS_REQUIRED = 2
    client_id = 'testuser'
    client_name = f'client.{client_id}'

    def setUp(self):
        super(SwapHelper, self).setUp()
        self.fs1 = self.fs
        self.fs2 = self.mds_cluster.newfs(name='testcephfs2', create=True)
        self.fs1.getinfo()
        self.fs2.getinfo()
        self.orig_fs_id_name = {self.fs1.id: self.fs1.name,
                                self.fs2.id: self.fs2.name}

        self.mount_a.remount(cephfs_name=self.fs1.name)
        self.mount_b.remount(cephfs_name=self.fs2.name)

        self.captesters = (CapTester(self.mount_a), CapTester(self.mount_b))

        self.mount_a.umount_wait()
        self.mount_b.umount_wait()

    def tearDown(self):
        self.mount_a.umount_wait()
        self.mount_b.umount_wait()
        self.run_ceph_cmd(args=f'auth rm {self.client_name}')
        super(SwapHelper, self).tearDown()

    def _reauthorize_client(self):
        moncap = gen_mon_cap_str((("rw", self.fs1.name),
                                  ("rw", self.fs2.name)))
        osdcap = gen_osd_cap_str((("rw", self.fs1.name),
                                  ("rw", self.fs2.name)))
        mdscap = gen_mds_cap_str((("rw", self.fs1.name),
                                  ("rw", self.fs2.name)))
        self.run_ceph_cmd(args=f'auth add {self.client_name} mon "{moncap}" '
                               f'osd "{osdcap}" mds "{mdscap}"')

    def _remount_both_cephfss(self):
        keyring = self.fs.mon_manager.get_keyring(self.client_id) + '\n'
        keyring_path_a = self.mount_a.client_remote.mktemp(data=keyring)
        keyring_path_b = self.mount_b.client_remote.mktemp(data=keyring)

        self.mount_a.mount(client_id=self.client_id,
                           client_keyring_path=keyring_path_a,
                           cephfs_mntpt='/', cephfs_name=self.fs1.name)
        self.mount_b.mount(client_id=self.client_id,
                           client_keyring_path=keyring_path_b,
                           cephfs_mntpt='/', cephfs_name=self.fs2.name)

    def run_rw_tests(self):
        for captester in self.captesters:
            captester.conduct_pos_test_for_read_caps()
            captester.conduct_pos_test_for_write_caps()

    def check_pool_application_metadata_key_value(self, pool, app, key, value):
        output = self.get_ceph_cmd_stdout(
            'osd', 'pool', 'application', 'get', pool, app, key)
        self.assertEqual(str(output.strip()), value)

    def _check_fs_name_on_fs_pools(self, fss):
        for fs in fss:
            for pool in fs.get_data_pool_names(True):
                self.check_pool_application_metadata_key_value(pool, 'cephfs',
                    'data', fs.name)
            self.check_pool_application_metadata_key_value(
                fs.get_metadata_pool_name(), 'cephfs', 'metadata', fs.name)

    def _are_fsnames_and_fscids_together(self):
        '''
        Are FS names and FSIDs together on same the FS as they were before
        running "ceph fs swap" command?
        '''
        fs1_id_swapped = self.orig_fs_id_name[self.fs1.id] == self.fs1.name
        fs2_id_swapped = self.orig_fs_id_name[self.fs2.id] == self.fs2.name

        if fs1_id_swapped and fs2_id_swapped:
            return True
        elif not fs1_id_swapped and not fs2_id_swapped:
            return False
        else:
            raise RuntimeError(
                'Unexpected situation occured: FSID for one FS swapped but '
                'not for the other FS.')

    def _bring_both_cephfss_down(self):
        self.run_ceph_cmd(f'fs fail {self.fs1.name}')
        self.run_ceph_cmd(f'fs fail {self.fs2.name}')

    def _bring_both_cephfss_up(self):
        self.run_ceph_cmd(f'fs set {self.fs1.name} joinable true')
        self.run_ceph_cmd(f'fs set {self.fs2.name} joinable true')

    def _refuse_clients_for_both_cephfss(self):
        self.run_ceph_cmd(f'fs set {self.fs1.name} refuse_client_session true')
        self.run_ceph_cmd(f'fs set {self.fs2.name} refuse_client_session true')

    def _accept_clients_for_both_cephfss(self):
        self.run_ceph_cmd(f'fs set {self.fs1.name} refuse_client_session '
                           'false')
        self.run_ceph_cmd(f'fs set {self.fs2.name} refuse_client_session '
                           'false')


class TestSwap(SwapHelper):
    '''
    Tests for "ceph fs swap" command.
    '''

    def test_swap_fsnames_but_not_fscids(self):
        '''
        Test that "ceph fs swap --swap-fscids=no" swaps the FS names but not
        the FSCIDs.
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        # log output to help debug test failures
        self.run_ceph_cmd('fs', 'dump')

        self.run_ceph_cmd(f'fs swap {self.fs1.name} {self.fs1.id} '
                          f'{self.fs2.name} {self.fs2.id} --swap-fscids=no '
                           '--yes_i_really_mean_it')

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # FS names were swapped but not FSIDs, so both can't be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), False)
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        self.run_rw_tests()

    def test_swap_fsnames_and_fscids(self):
        '''
        Test that "ceph fs swap --swap-fscids=yes" swaps the FS names as well
        as the FSCIDs.
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.run_ceph_cmd(f'fs swap {self.fs1.name} {self.fs1.id} '
                          f'{self.fs2.name} {self.fs2.id} --swap-fscids=yes '
                           '--yes_i_really_mean_it')

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        # XXX: Let's ensure that FS mounted on a mountpoint is same before
        # and after swapping of FS name and FSCIDs. This ensures that data
        # available on mountpoints before and after the swap is same. This
        # prevents self.run_rw_tests() from breaking.
        #
        # At the beginning of test, testcephfs1 has data (let's say) 'abc1'
        # and testcephfs2 has data 'abc2'. self.fs1 is mapped to testcephfs1
        # and self.fs2 mapped to testcephfs2. After swap, data of testcephfs1
        # and testcephfs2 will be 'abc2' and 'abc1' respectively.
        #
        # However, running self.fs1.getinfo() after swap will map self.fs1 to
        # FS with FSCID 1 i.e. testcephfs1 and not testcephfs2. Thus, data
        # under self.fs1 will be different than data before swapping. This
        # breaks self.run_rw_tests() because self.fs1 is always mounted on
        # the mountpoint of self.mount_a.

        # To prevent this, therefore, make sure that data on
        # self.fs1/self.mount_a is same after and before the swap. To ensure
        # this, swap FS that is represented by self.fs1. Instead of
        # testcephfs1 it should be testcephfs2 because, after swap,
        # testcephfs2 containts the data of testcephfs1. This will ensure that
        # self.mount_rw_tests doesn't break.
        #
        # Same for self.fs2.
        self.fs1.id, self.fs2.id = None, None
        self.fs1.name, self.fs2.name = self.fs2.name, self.fs1.name
        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # both FS name and FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        self.run_rw_tests()

    def test_swap_without_confirmation_option(self):
        '''
        Test that "ceph fs swap --swap-fscids=yes" without the option
        "--yes-i-really-mean-it" fails.
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = ('This is a potentially disruptive operation, client\'s cephx '
               'credentials may need to be reauthorized to access the file '
               'systems and its pools. Add --yes-i-really-mean-it if you are '
               'sure you wish to continue.')
        self.negtest_ceph_cmd(f'fs swap {self.fs1.name} {self.fs1.id} '
                              f'{self.fs2.name} {self.fs2.id} '
                               '--swap-fscids=no',
            retval=errno.EPERM, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)


class TestSwapFsAbsent(SwapHelper):
    '''
    Tests for "fs swap" when either FS name is false.
    '''

    def test_swap_when_fs1_is_absent(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when there is no CephFS on cluster by the name "<fs1name>".
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        absent_cephfs = 'random_fsname_654'
        msg = (f"File system '{absent_cephfs}' doesn't exist on this Ceph "
                "cluster")
        self.negtest_ceph_cmd(
            args=(f'fs swap {absent_cephfs} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.ENOENT, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_fs2_is_absent(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when there is no CephFS on cluster by the name "<fs2name>".
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        absent_cephfs = 'random_fsname_654'
        msg = (f"File system '{absent_cephfs}' doesn't exist on this Ceph "
                "cluster")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs2.id} {absent_cephfs} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.ENOENT, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_both_fss_are_absent(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when there are no CephFSs on the cluster by the name "<fs1name>" and
        "<fs2name>".
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        absent_cephfs1 = 'random_fsname_65'
        absent_cephfs2 = 'random_fsname_66'
        msg = (f"Neither file system '{absent_cephfs1}' nor file system "
               f"'{absent_cephfs2}' exists on this Ceph cluster")
        self.negtest_ceph_cmd(
            args=(f'fs swap {absent_cephfs1} 123 {absent_cephfs2} 1234 '
                   '--swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.ENOENT, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)


class TestSwapFscidWrong(SwapHelper):
    '''
    Tests for "fs swap" when either FSCID is wrong.
    '''

    def test_swap_when_fs1_id_is_wrong(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when "<fs1id>" is not the FSCID of the CephFS named "<fs1nmae>".
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"FSCID provided for '{self.fs1.name}' is incorrect.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} 123 {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EINVAL, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_fs2_id_is_wrong(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when "<fs2id>" is not the FSCID of the CephFS named "<fs2nmae>".
        '''
        self._bring_both_cephfss_down()
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"FSCID provided for '{self.fs2.name}' is incorrect.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'123 --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EINVAL, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_both_fscids_are_wrong(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when "<fs1id>" and "<fs2id>", respectively, are not the FSCIDs of the
        CephFSs named "<fs1name>" and "<fs2nmae>".
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = ('FSCIDs provided for both the CephFSs is incorrect.')
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} 123 {self.fs2.name} 1234 '
                  f'--swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EINVAL, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_user_swaps_fscids_in_cmd_args(self):
        '''
        Test that "ceph fs swap" fails and prints relevant error message when
        FSCIDs are exchange while writing the command. That is user write the
        command as -

        "ceph fs swap <fs1name> <fs2id> <fs2name> <fs1id>"

        instead of writing -

        "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>"
        '''
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = ('FSCIDs provided in command arguments are swapped; perhaps '
               '`ceph fs swap` has been run before.')
        proc = self.run_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs2.id} {self.fs2.name} '
                  f'{self.fs1.id} --swap-fscids=no --yes_i_really_mean_it'),
            stderr=StringIO())
        self.assertIn(msg.lower(), proc.stderr.getvalue().lower())

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)


class TestSwapMirroringOn(SwapHelper):
    '''
    # Tests for "fs swap" when mirroring is enabled on FS
    '''

    def test_swap_when_mirroring_enabled_for_1st_FS(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when mirroring is enabled for the CephFS named "<fs1name>".
        '''
        self.run_ceph_cmd(f'fs mirror enable {self.fs1.name}')
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"Mirroring is enabled on file system '{self.fs1.name}'. "
                "Disable mirroring on the file system after ensuring it's "
                "OK to do so, and then re-try swapping.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')
        self.run_ceph_cmd(f'fs mirror disable {self.fs1.name}')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_mirroring_enabled_for_2nd_FS(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when mirroring is enabled for the CephFS named "<fs2name>".
        '''
        self.run_ceph_cmd(f'fs mirror enable {self.fs2.name}')
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)

        self.run_ceph_cmd('fs', 'dump')
        msg = (f"Mirroring is enabled on file system '{self.fs2.name}'. "
                "Disable mirroring on the file system after ensuring it's "
                "OK to do so, and then re-try swapping.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')
        self.run_ceph_cmd(f'fs mirror disable {self.fs2.name}')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_mirroring_enabled_for_both_FSs(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" fails
        when mirroring is enabled for both the CephFSs.
        '''
        self.run_ceph_cmd(f'fs mirror enable {self.fs1.name}')
        self.run_ceph_cmd(f'fs mirror enable {self.fs2.name}')
        self._bring_both_cephfss_down()
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"Mirroring is enabled on file systems '{self.fs1.name}' and "
               f"'{self.fs2.name}'. Disable mirroring on both the file "
                "systems after ensuring it's OK to do so, and then re-try "
                "swapping.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self._bring_both_cephfss_up()
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')
        self.run_ceph_cmd(f'fs mirror disable {self.fs1.name}')
        self.run_ceph_cmd(f'fs mirror disable {self.fs2.name}')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)


class TestSwapFsOnline(SwapHelper):
    '''
    Tests for "fs swap" when either FS is not down/failed.
    '''

    def test_swap_when_fs1_is_online(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" when
        CephFS named "<fs1name>" is online (i.e. is not failed).
        '''
        self.run_ceph_cmd(f'fs fail {self.fs2.name}')
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"CephFS '{self.fs1.name}' is not offline. Before swapping "
                "CephFS names, both CephFSs should be marked as failed. "
                "See `ceph fs fail`.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self.run_ceph_cmd(f'fs set {self.fs2.name} joinable true')
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_fs2_is_online(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" when
        CephFS named "<fs2name>" is online (i.e. is not failed).
        '''
        self.run_ceph_cmd(f'fs fail {self.fs1.name}')
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"CephFS '{self.fs2.name}' is not offline. Before swapping "
                "CephFS names, both CephFSs should be marked as failed. "
                "See `ceph fs fail`.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self.run_ceph_cmd(f'fs set {self.fs1.name} joinable true')
        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_both_FSs_are_online(self):
        '''
        Test that "ceph fs swap <fs1name> <fs1id> <fs2name> <fs2id>" when
        both the CephFSs are online (i.e. is not failed).
        '''
        self._refuse_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"CephFSs '{self.fs1.name}' and '{self.fs2.name}' are not "
                "offline. Before swapping CephFS names, both CephFSs should "
                "be marked as failed. See `ceph fs fail`.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self._accept_clients_for_both_cephfss()
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)


class TestSwapDoesntRefuseClients(SwapHelper):
    '''
    Tests for "fs swap" when either FS is offline.
    '''

    def test_swap_when_FS1_doesnt_refuse_clients(self):
        '''
        Test that the command "ceph fs swap" command fails when
        "refuse_client_session" is not set for the first of the two of FSs .
        '''
        self._bring_both_cephfss_down()
        self.run_ceph_cmd(f'fs set {self.fs2.name} refuse_client_session true')
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"CephFS '{self.fs1.name}' doesn't refuse clients. Before "
                "swapping CephFS names, flag 'refuse_client_session' must "
                "be set. See `ceph fs set`.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self._bring_both_cephfss_up()
        self.run_ceph_cmd(f'fs set {self.fs2.name} refuse_client_session '
                           'false')
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_FS2_doesnt_refuse_clients(self):
        '''
        Test that the command "ceph fs swap" command fails when
        "refuse_client_session" is not set for the second of the two of FSs .
        '''
        self._bring_both_cephfss_down()
        self.run_ceph_cmd(f'fs set {self.fs1.name} refuse_client_session true')
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        msg = (f"CephFS '{self.fs2.name}' doesn't refuse clients. Before "
                "swapping CephFS names, flag 'refuse_client_session' must "
                "be set. See `ceph fs set`.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)

        self._bring_both_cephfss_up()
        self.run_ceph_cmd(f'fs set {self.fs1.name} refuse_client_session '
                           'false')
        sleep(2)
        self.run_ceph_cmd('fs', 'dump')

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)

    def test_swap_when_both_FSs_do_not_refuse_clients(self):
        '''
        Test that the command "ceph fs swap" command fails when
        "refuse_client_session" is not set for both the CephFSs.
        '''
        self.run_ceph_cmd('fs', 'dump')
        self._bring_both_cephfss_down()
        sleep(2)
        msg = (f"CephFSs '{self.fs1.name}' and '{self.fs2.name}' do not "
                "refuse clients. Before swapping CephFS names, flag "
                "'refuse_client_session' must be set. See `ceph fs set`.")
        self.negtest_ceph_cmd(
            args=(f'fs swap {self.fs1.name} {self.fs1.id} {self.fs2.name} '
                  f'{self.fs2.id} --swap-fscids=no --yes_i_really_mean_it'),
            retval=errno.EPERM, errmsgs=msg)
        self._bring_both_cephfss_up()
        self.run_ceph_cmd('fs', 'dump')
        sleep(2)

        self.fs1.getinfo()
        self.fs2.getinfo()
        self._reauthorize_client()
        self._remount_both_cephfss()

        # check that content of both CephFSs is unaffected by this failure.
        self.run_rw_tests()
        self._check_fs_name_on_fs_pools((self.fs1, self.fs2))
        # neither FS name nor FSIDs were swapped, so both must be together
        self.assertEqual(self._are_fsnames_and_fscids_together(), True)




#
# Copyright(c) 2019 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause-Clear
#

from ctypes import c_uint32, c_uint64, Structure

from .shared import OcfStatsReq, OcfStatsBlock, OcfStatsDebug, OcfStatsError


class CoreInfo(Structure):
    _fields_ = [
        ("core_size", c_uint64),
        ("core_size_bytes", c_uint64),
        ("dirty", c_uint32),
        ("flushed", c_uint32),
        ("dirty_for", c_uint32),
        ("seq_cutoff_threshold", c_uint32),
        ("seq_cutoff_policy", c_uint32),
    ]



from .helpers import deprecated_alias


@deprecated_alias('add_ip_address')
def net_interface_add_ip_address(client, ifc_index, ip_addr):
    """Add IP address.

    Args:
        ifc_index: ifc index of the nic device (int)
        ip_addr: ip address will be added
    """
    params = {'ifc_index': ifc_index, 'ip_address': ip_addr}
    return client.call('net_interface_add_ip_address', params)


@deprecated_alias('delete_ip_address')
def net_interface_delete_ip_address(client, ifc_index, ip_addr):
    """Delete IP address.

    Args:
        ifc_index: ifc index of the nic device (int)
        ip_addr: ip address will be deleted
    """
    params = {'ifc_index': ifc_index, 'ip_address': ip_addr}
    return client.call('net_interface_delete_ip_address', params)


@deprecated_alias('get_interfaces')
def net_get_interfaces(client):
    """Display current interface list

    Returns:
        List of current interface
    """
    return client.call('net_get_interfaces')



"""
test_stress_watch task
"""
import contextlib
import logging

from teuthology.orchestra import run
from teuthology.task import proc_thrasher

log = logging.getLogger(__name__)


@contextlib.contextmanager
def task(ctx, config):
    """
    Run test_stress_watch

    The config should be as follows:

    test_stress_watch:
        clients: [client list]

    example:

    tasks:
    - ceph:
    - test_stress_watch:
        clients: [client.0]
    - interactive:
    """
    log.info('Beginning test_stress_watch...')
    assert isinstance(config, dict), \
        "please list clients to run on"
    testwatch = {}

    remotes = []

    for role in config.get('clients', ['client.0']):
        assert isinstance(role, str)
        PREFIX = 'client.'
        assert role.startswith(PREFIX)
        id_ = role[len(PREFIX):]
        (remote,) = ctx.cluster.only(role).remotes.keys()
        remotes.append(remote)

        args =['CEPH_CLIENT_ID={id_}'.format(id_=id_),
               'CEPH_ARGS="{flags}"'.format(flags=config.get('flags', '')),
               'daemon-helper',
               'kill',
               'multi_stress_watch foo foo'
               ]

        log.info("args are %s" % (args,))

        proc = proc_thrasher.ProcThrasher({}, remote,
            args=[run.Raw(i) for i in args],
            logger=log.getChild('testwatch.{id}'.format(id=id_)),
            stdin=run.PIPE,
            wait=False
            )
        proc.start()
        testwatch[id_] = proc

    try:
        yield
    finally:
        log.info('joining watch_notify_stress')
        for i in testwatch.values():
            i.join()



"""
Force pg creation on all osds
"""
from teuthology import misc as teuthology
from teuthology.orchestra import run
import logging

log = logging.getLogger(__name__)

def task(ctx, config):
    """
    Create the specified number of pools and write 16 objects to them (thereby forcing
    the PG creation on each OSD). This task creates pools from all the clients,
    in parallel. It is easy to add other daemon types which have the appropriate
    permissions, but I don't think anything else does.
    The config is just the number of pools to create. I recommend setting
    "mon create pg interval" to a very low value in your ceph config to speed
    this up.
    
    You probably want to do this to look at memory consumption, and
    maybe to test how performance changes with the number of PGs. For example:
    
    tasks:
    - ceph:
        config:
          mon:
            mon create pg interval: 1
    - manypools: 3000
    - radosbench:
        clients: [client.0]
        time: 360
    """
    
    log.info('creating {n} pools'.format(n=config))
    
    poolnum = int(config)
    creator_remotes = []
    client_roles = teuthology.all_roles_of_type(ctx.cluster, 'client')
    log.info('got client_roles={client_roles_}'.format(client_roles_=client_roles))
    for role in client_roles:
        log.info('role={role_}'.format(role_=role))
        (creator_remote, ) = ctx.cluster.only('client.{id}'.format(id=role)).remotes.keys()
        creator_remotes.append((creator_remote, 'client.{id}'.format(id=role)))

    remaining_pools = poolnum
    poolprocs=dict()
    while (remaining_pools > 0):
        log.info('{n} pools remaining to create'.format(n=remaining_pools))
        for remote, role_ in creator_remotes:
            poolnum = remaining_pools
            remaining_pools -= 1
            if remaining_pools < 0:
                continue
            log.info('creating pool{num} on {role}'.format(num=poolnum, role=role_))
            proc = remote.run(
                args=[
                    'ceph',
                    '--name', role_,
                    'osd', 'pool', 'create', 'pool{num}'.format(num=poolnum), '8',
                    run.Raw('&&'),
                    'rados',
                    '--name', role_,
                    '--pool', 'pool{num}'.format(num=poolnum),
                    'bench', '0', 'write', '-t', '16', '--block-size', '1'
                ],
                wait = False
            )
            log.info('waiting for pool and object creates')
            poolprocs[remote] = proc

        run.wait(poolprocs.values())

    log.info('created all {n} pools and wrote 16 objects to each'.format(n=poolnum))



"""
Functions to netsplit test machines.

At present, you must specify monitors to disconnect, and it
drops those IP pairs. This means OSDs etc on the hosts which use
the same IP will also be blocked! If you are using multiple IPs on the
same host within the cluster, daemons on those other IPs will get
through.
"""
import logging
import re

log = logging.getLogger(__name__)


def get_ip_and_ports(ctx, daemon):
    """
    Get the IP and port list for the <daemon>.
    """
    assert daemon.startswith('mon.')
    addr = ctx.ceph['ceph'].mons['{a}'.format(a=daemon)]
    ips = re.findall("[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+[:[0-9]*]*", addr)
    assert len(ips) > 0
    plain_ip = re.match("[0-9\.]*", ips[0]).group()
    assert plain_ip is not None
    port_list = []
    for ip in ips:
        ip_str, port_str = re.match("([0-9\.]*)([:[0-9]*]*)", ip).groups()
        assert ip_str == plain_ip
        if len(port_str) > 0:
            port_list.append(port_str)
    return (plain_ip, port_list)


def disconnect(ctx, config):
    """
    Disconnect the mons in the <config> list.
    """
    assert len(config) == 2  # we can only disconnect pairs right now
    # and we can only disconnect mons right now
    assert config[0].startswith('mon.')
    assert config[1].startswith('mon.')
    log.info("Disconnecting {a} and {b}".format(a=config[0], b=config[1]))
    (ip1, _) = get_ip_and_ports(ctx, config[0])
    (ip2, _) = get_ip_and_ports(ctx, config[1])

    (host1,) = ctx.cluster.only(config[0]).remotes.keys()
    (host2,) = ctx.cluster.only(config[1]).remotes.keys()
    assert host1 is not None
    assert host2 is not None

    host1.run(args=["sudo", "iptables", "-A", "INPUT",
                    "-s", ip2, "-j", "DROP"])
    host1.run(args=["sudo", "iptables", "-A", "OUTPUT",
                    "-d", ip2, "-j", "DROP"])

    host2.run(args=["sudo", "iptables", "-A", "INPUT",
                    "-s", ip1, "-j", "DROP"])
    host2.run(args=["sudo", "iptables", "-A", "OUTPUT",
                    "-d", ip1, "-j", "DROP"])


def reconnect(ctx, config):
    """
    Reconnect the mons in the <config> list.
    """
    assert len(config) == 2  # we can only disconnect pairs right now
    # and we can only disconnect mons right now
    assert config[0].startswith('mon.')
    assert config[1].startswith('mon.')
    log.info("Reconnecting {a} and {b}".format(a=config[0], b=config[1]))
    (ip1, _) = get_ip_and_ports(ctx, config[0])
    (ip2, _) = get_ip_and_ports(ctx, config[1])

    (host1,) = ctx.cluster.only(config[0]).remotes.keys()
    (host2,) = ctx.cluster.only(config[1]).remotes.keys()
    assert host1 is not None
    assert host2 is not None

    host1.run(args=["sudo", "iptables", "-D", "INPUT",
                    "-s", ip2, "-j", "DROP"])
    host1.run(args=["sudo", "iptables", "-D", "OUTPUT",
                    "-d", ip2, "-j", "DROP"])

    host2.run(args=["sudo", "iptables", "-D", "INPUT",
                    "-s", ip1, "-j", "DROP"])
    host2.run(args=["sudo", "iptables", "-D", "OUTPUT",
                    "-d", ip1, "-j", "DROP"])



#!/usr/bin/env python3
import sys
import argparse
from configshell_fb import ConfigShell, shell, ExecutionError
from pyparsing import (alphanums, Optional, Suppress, Word, Regex,
                       removeQuotes, dblQuotedString, OneOrMore)
from rpc.client import JSONRPCException, JSONRPCClient
from spdkcli import UIRoot


def add_quotes_to_shell(spdk_shell):
    command = shell.locatedExpr(Word(alphanums + '_'))('command')
    value = dblQuotedString.addParseAction(removeQuotes)
    value_word = Word(alphanums + r';,=_\+/.<>()~@:-%[]')
    keyword = Word(alphanums + r'_\-')
    kparam = shell.locatedExpr(keyword + Suppress('=') +
                               Optional(value | value_word, default=''))('kparams*')
    pparam = shell.locatedExpr(value | value_word)('pparams*')
    parameters = OneOrMore(kparam | pparam)
    bookmark = Regex(r'@([A-Za-z0-9:_.]|-)+')
    pathstd = Regex(r'([A-Za-z0-9:_.\[\]]|-)*' + '/' + r'([A-Za-z0-9:_.\[\]/]|-)*') \
        | '..' | '.'
    path = shell.locatedExpr(bookmark | pathstd | '*')('path')
    spdk_shell._parser = Optional(path) + Optional(command) + Optional(parameters)


def main():
    """
    Start SPDK CLI
    :return:
    """
    spdk_shell = ConfigShell("~/.scripts")
    spdk_shell.interactive = True
    add_quotes_to_shell(spdk_shell)

    parser = argparse.ArgumentParser(description="SPDK command line interface")
    parser.add_argument('-s', dest='server_addr',
                        help='RPC domain socket path or IP address', default='/var/tmp/spdk.sock')
    parser.add_argument('-p', dest='port',
                        help='RPC port number (if server_addr is IP address)',
                        default=None, type=int)
    parser.add_argument("-v", dest="verbose", help="Print request/response JSON for configuration calls",
                        default=False, action="store_true")
    parser.add_argument("commands", metavar="command", type=str, nargs="*", default="",
                        help="commands to execute by SPDKCli as one-line command")
    args = parser.parse_args()

    try:
        client = JSONRPCClient(args.server_addr, port=args.port)
    except JSONRPCException as e:
        spdk_shell.log.error("%s. SPDK not running?" % e)
        sys.exit(1)

    with client:
        root_node = UIRoot(client, spdk_shell)
        root_node.verbose = args.verbose
        try:
            root_node.refresh()
        except BaseException:
            pass

        if args.commands:
            try:
                spdk_shell.interactive = False
                spdk_shell.run_cmdline(" ".join(args.commands))
            except Exception as e:
                sys.stderr.write("%s\n" % e)
                sys.exit(1)
            sys.exit(0)

        spdk_shell.con.display("SPDK CLI v0.1")
        spdk_shell.con.display("")

        while not spdk_shell._exit:
            try:
                spdk_shell.run_interactive()
            except (JSONRPCException, ExecutionError) as e:
                spdk_shell.log.error("%s" % e)
            except BrokenPipeError as e:
                spdk_shell.log.error("Lost connection with SPDK: %s" % e)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

import requests
from urllib3.exceptions import MaxRetryError

from .helper import (DashboardTestCase, JLeaf, JList, JObj,
                     module_options_object_schema, module_options_schema,
                     retry)

logger = logging.getLogger(__name__)


class MgrModuleTestCase(DashboardTestCase):
    MGRS_REQUIRED = 1

    @retry(on_exception=RuntimeError, tries=2, delay=0.5, logger=logger)
    def wait_until_rest_api_accessible(self):
        """
        Wait until the REST API is accessible.
        """

        def _check_connection():
            try:
                # Try reaching an API endpoint successfully.
                logger.info('Trying to reach the REST API endpoint')
                self._get('/api/mgr/module')
                if self._resp.status_code == 200:
                    return True
            except (MaxRetryError, requests.ConnectionError):
                pass
            return False

        self.wait_until_true(_check_connection, timeout=30)


class MgrModuleTest(MgrModuleTestCase):

    def test_list_disabled_module(self):
        self._ceph_cmd(['mgr', 'module', 'disable', 'iostat'], wait=3)
        data = self._get(
            '/api/mgr/module',
            retries=1,
            wait_func=lambda:  # pylint: disable=unnecessary-lambda
            self.wait_until_rest_api_accessible()
        )
        self.assertStatus(200)
        self.assertSchema(
            data,
            JList(
                JObj(sub_elems={
                    'name': JLeaf(str),
                    'enabled': JLeaf(bool),
                    'always_on': JLeaf(bool),
                    'options': module_options_schema
                })))
        module_info = self.find_object_in_list('name', 'iostat', data)
        self.assertIsNotNone(module_info)
        self.assertFalse(module_info['enabled'])

    def test_list_enabled_module(self):
        self._ceph_cmd(['mgr', 'module', 'enable', 'iostat'], wait=3)
        data = self._get(
            '/api/mgr/module',
            retries=1,
            wait_func=lambda:  # pylint: disable=unnecessary-lambda
            self.wait_until_rest_api_accessible()
        )
        self.assertStatus(200)
        self.assertSchema(
            data,
            JList(
                JObj(sub_elems={
                    'name': JLeaf(str),
                    'enabled': JLeaf(bool),
                    'always_on': JLeaf(bool),
                    'options': module_options_schema
                })))
        module_info = self.find_object_in_list('name', 'iostat', data)
        self.assertIsNotNone(module_info)
        self.assertTrue(module_info['enabled'])

    def test_get(self):
        data = self._get('/api/mgr/module/telemetry')
        self.assertStatus(200)
        self.assertSchema(
            data,
            JObj(
                allow_unknown=True,
                sub_elems={
                    'channel_basic': bool,
                    'channel_ident': bool,
                    'channel_crash': bool,
                    'channel_device': bool,
                    'channel_perf': bool,
                    'contact': str,
                    'description': str,
                    'enabled': bool,
                    'interval': int,
                    'last_opt_revision': int,
                    'leaderboard': bool,
                    'leaderboard_description': str,
                    'organization': str,
                    'proxy': str,
                    'url': str
                }))

    def test_module_options(self):
        data = self._get('/api/mgr/module/telemetry/options')
        self.assertStatus(200)
        schema = JObj({
            'channel_basic': module_options_object_schema,
            'channel_crash': module_options_object_schema,
            'channel_device': module_options_object_schema,
            'channel_ident': module_options_object_schema,
            'channel_perf': module_options_object_schema,
            'contact': module_options_object_schema,
            'description': module_options_object_schema,
            'device_url': module_options_object_schema,
            'enabled': module_options_object_schema,
            'interval': module_options_object_schema,
            'last_opt_revision': module_options_object_schema,
            'leaderboard': module_options_object_schema,
            'leaderboard_description': module_options_object_schema,
            'sqlite3_killpoint': module_options_object_schema,
            'log_level': module_options_object_schema,
            'log_to_cluster': module_options_object_schema,
            'log_to_cluster_level': module_options_object_schema,
            'log_to_file': module_options_object_schema,
            'organization': module_options_object_schema,
            'proxy': module_options_object_schema,
            'url': module_options_object_schema
        })
        self.assertSchema(data, schema)

    def test_module_enable(self):
        self._post('/api/mgr/module/telemetry/enable')
        self.assertStatus(200)

    def test_disable(self):
        self._post('/api/mgr/module/iostat/disable')
        self.assertStatus(200)

    def test_put(self):
        self.set_config_key('config/mgr/mgr/iostat/log_level', 'critical')
        self.set_config_key('config/mgr/mgr/iostat/log_to_cluster', 'False')
        self.set_config_key('config/mgr/mgr/iostat/log_to_cluster_level', 'info')
        self.set_config_key('config/mgr/mgr/iostat/log_to_file', 'True')
        self._put(
            '/api/mgr/module/iostat',
            data={
                'config': {
                    'log_level': 'debug',
                    'log_to_cluster': True,
                    'log_to_cluster_level': 'warning',
                    'log_to_file': False
                }
            })
        self.assertStatus(200)
        data = self._get('/api/mgr/module/iostat')
        self.assertStatus(200)
        self.assertEqual(data['log_level'], 'debug')
        self.assertTrue(data['log_to_cluster'])
        self.assertEqual(data['log_to_cluster_level'], 'warning')
        self.assertFalse(data['log_to_file'])



"""
Aver wrapper task
"""
import contextlib
import logging
from subprocess import check_call, Popen, PIPE

log = logging.getLogger(__name__)


@contextlib.contextmanager
def task(ctx, config):
    """
    Execute an aver assertion

    Parameters:

        input: file containing data referred to by the assertions. File name is
               relative to the job's archive path
        validations: list of validations in the Aver language

    Example:
    - aver:
        input: bench_output.csv
        validations:
        - expect performance(alg='ceph') > performance(alg='raw')
        - for size > 3 expect avg_throughput > 2000
    """
    log.info('Beginning aver...')
    assert isinstance(config, dict), 'expecting dictionary for configuration'

    if 'input' not in config:
        raise Exception("Expecting 'input' option")
    if len(config.get('validations', [])) < 1:
        raise Exception("Expecting at least one entry in 'validations'")

    url = ('https://github.com/ivotron/aver/releases/download/'
           'v0.3.0/aver-linux-amd64.tar.bz2')

    aver_path = ctx.archive + '/aver'

    # download binary
    check_call(['wget', '-O', aver_path + '.tbz', url])
    check_call(['tar', 'xfj', aver_path + '.tbz', '-C', ctx.archive])

    # print version
    process = Popen([aver_path, '-v'], stdout=PIPE)
    log.info(process.communicate()[0])

    # validate
    for validation in config['validations']:
        cmd = (aver_path + ' -s -i ' + (ctx.archive + '/' + config['input']) +
               ' "' + validation + '"')
        log.info("executing: " + cmd)
        process = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        (stdout, stderr) = process.communicate()
        if stderr:
            log.info('aver stderr: ' + stderr)
        log.info('aver result: ' + stdout)
        if stdout.strip(' \t\n\r') != 'true':
            raise Exception('Failed validation: ' + validation)

    try:
        yield
    finally:
        log.info('Removing aver binary...')
        check_call(['rm', aver_path, aver_path + '.tbz'])



#
# Copyright(c) 2019 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause-Clear
#

import logging
from ctypes import CFUNCTYPE, c_size_t, c_char_p, Structure, c_void_p
from enum import IntEnum, auto
from threading import Event

from ..utils import Size as S


class OcfErrorCode(IntEnum):
    OCF_ERR_INVAL = 1000000
    OCF_ERR_AGAIN = auto()
    OCF_ERR_INTR = auto()
    OCF_ERR_NOT_SUPP = auto()
    OCF_ERR_NO_MEM = auto()
    OCF_ERR_NO_LOCK = auto()
    OCF_ERR_METADATA_VER = auto()
    OCF_ERR_NO_METADATA = auto()
    OCF_ERR_METADATA_FOUND = auto()
    OCF_ERR_INVAL_VOLUME_TYPE = auto()
    OCF_ERR_UNKNOWN = auto()
    OCF_ERR_TOO_MANY_CACHES = auto()
    OCF_ERR_NO_FREE_RAM = auto()
    OCF_ERR_START_CACHE_FAIL = auto()
    OCF_ERR_CACHE_NOT_EXIST = auto()
    OCF_ERR_CORE_NOT_EXIST = auto()
    OCF_ERR_CACHE_EXIST = auto()
    OCF_ERR_CORE_EXIST = auto()
    OCF_ERR_TOO_MANY_CORES = auto()
    OCF_ERR_CORE_NOT_AVAIL = auto()
    OCF_ERR_NOT_OPEN_EXC = auto()
    OCF_ERR_CACHE_NOT_AVAIL = auto()
    OCF_ERR_IO_CLASS_NOT_EXIST = auto()
    OCF_ERR_IO = auto()
    OCF_ERR_WRITE_CACHE = auto()
    OCF_ERR_WRITE_CORE = auto()
    OCF_ERR_DIRTY_SHUTDOWN = auto()
    OCF_ERR_DIRTY_EXISTS = auto()
    OCF_ERR_FLUSHING_INTERRUPTED = auto()
    OCF_ERR_FLUSH_IN_PROGRESS = auto()
    OCF_ERR_CANNOT_ADD_CORE_TO_POOL = auto()
    OCF_ERR_CACHE_IN_INCOMPLETE_STATE = auto()
    OCF_ERR_CORE_IN_INACTIVE_STATE = auto()
    OCF_ERR_INVALID_CACHE_MODE = auto()
    OCF_ERR_INVALID_CACHE_LINE_SIZE = auto()
    OCF_ERR_CACHE_NAME_MISMATCH = auto()
    OCF_ERR_INVAL_CACHE_DEV = auto()


class OcfCompletion:
    """
    This class provides Completion mechanism for interacting with OCF async
    management API.
    """

    class CompletionResult:
        def __init__(self, completion_args):
            self.completion_args = {
                x[0]: i for i, x in enumerate(completion_args)
            }
            self.results = None
            self.arg_types = [x[1] for x in completion_args]

        def __getitem__(self, key):
            try:
                position = self.completion_args[key]
                return self.results[position]
            except KeyError:
                raise KeyError(f"No completion argument {key} specified")

    def __init__(self, completion_args: list):
        """
        Provide ctypes arg list, and optionally index of status argument in
        completion function which will be extracted (default - last argument).

        :param completion_args: list of tuples (parameter name, parameter type)
            for OCF completion function
        """
        self.e = Event()
        self.results = OcfCompletion.CompletionResult(completion_args)
        self._as_parameter_ = self.callback

    @property
    def callback(self):
        @CFUNCTYPE(c_void_p, *self.results.arg_types)
        def complete(*args):
            self.results.results = args
            self.e.set()

        return complete

    def wait(self):
        self.e.wait()


class OcfError(BaseException):
    def __init__(self, msg, error_code):
        super().__init__(self, msg)
        self.error_code = OcfErrorCode(abs(error_code))
        self.msg = msg

    def __str__(self):
        return "{} ({})".format(self.msg, repr(self.error_code))


class SharedOcfObject(Structure):
    _instances_ = {}

    def __init__(self):
        super().__init__()
        type(self)._instances_[self._as_parameter_] = self

    @classmethod
    def get_instance(cls, ref: int):
        try:
            return cls._instances_[ref]
        except:  # noqa E722
            logging.getLogger("pyocf").error(
                "OcfSharedObject corruption. wanted: {} instances: {}".format(
                    ref, cls._instances_
                )
            )
            return None

    @classmethod
    def del_object(cls, ref: int):
        del cls._instances_[ref]


class Uuid(Structure):
    _fields_ = [("_size", c_size_t), ("_data", c_char_p)]


class CacheLineSize(IntEnum):
    LINE_4KiB = S.from_KiB(4)
    LINE_8KiB = S.from_KiB(8)
    LINE_16KiB = S.from_KiB(16)
    LINE_32KiB = S.from_KiB(32)
    LINE_64KiB = S.from_KiB(64)
    DEFAULT = LINE_4KiB


class SeqCutOffPolicy(IntEnum):
    ALWAYS = 0
    FULL = 1
    NEVER = 2
    DEFAULT = FULL


class CacheLines(S):
    def __init__(self, count: int, line_size: CacheLineSize):
        self.bytes = count * line_size
        self.line_size = line_size

    def __int__(self):
        return int(self.bytes / self.line_size)



from __future__ import absolute_import

from .helper import DashboardTestCase


class ClusterConfigurationTest(DashboardTestCase):

    def test_list(self):
        data = self._get('/api/cluster_conf')
        self.assertStatus(200)
        self.assertIsInstance(data, list)
        self.assertGreater(len(data), 1000)
        for conf in data:
            self._validate_single(conf)

    def test_get(self):
        data = self._get('/api/cluster_conf/admin_socket')
        self.assertStatus(200)
        self._validate_single(data)
        self.assertIn('enum_values', data)

        data = self._get('/api/cluster_conf/fantasy_name')
        self.assertStatus(404)

    def test_get_specific_db_config_option(self):
        config_name = 'mon_allow_pool_delete'

        orig_value = self._get_config_by_name(config_name)

        self._ceph_cmd(['config', 'set', 'mon', config_name, 'true'])
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            [{'section': 'mon', 'value': 'true'}],
            timeout=30,
            period=1)

        self._ceph_cmd(['config', 'set', 'mon', config_name, 'false'])
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            [{'section': 'mon', 'value': 'false'}],
            timeout=30,
            period=1)

        # restore value
        if orig_value:
            self._ceph_cmd(['config', 'set', 'mon', config_name, orig_value[0]['value']])

    def test_filter_config_options(self):
        config_names = ['osd_scrub_during_recovery', 'osd_scrub_begin_hour', 'osd_scrub_end_hour']
        data = self._get('/api/cluster_conf/filter?names={}'.format(','.join(config_names)))
        self.assertStatus(200)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 3)
        for conf in data:
            self._validate_single(conf)
            self.assertIn(conf['name'], config_names)

    def test_filter_config_options_empty_names(self):
        self._get('/api/cluster_conf/filter?names=')
        self.assertStatus(404)
        self.assertEqual(self._resp.json()['detail'], 'Config options `` not found')

    def test_filter_config_options_unknown_name(self):
        self._get('/api/cluster_conf/filter?names=abc')
        self.assertStatus(404)
        self.assertEqual(self._resp.json()['detail'], 'Config options `abc` not found')

    def test_filter_config_options_contains_unknown_name(self):
        config_names = ['osd_scrub_during_recovery', 'osd_scrub_begin_hour', 'abc']
        data = self._get('/api/cluster_conf/filter?names={}'.format(','.join(config_names)))
        self.assertStatus(200)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 2)
        for conf in data:
            self._validate_single(conf)
            self.assertIn(conf['name'], config_names)

    def test_create(self):
        config_name = 'debug_ms'
        orig_value = self._get_config_by_name(config_name)

        # remove all existing settings for equal preconditions
        self._clear_all_values_for_config_option(config_name)

        expected_result = [{'section': 'mon', 'value': '0/3'}]

        self._post('/api/cluster_conf', {
            'name': config_name,
            'value': expected_result
        })
        self.assertStatus(201)
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            expected_result,
            timeout=30,
            period=1)

        # reset original value
        self._clear_all_values_for_config_option(config_name)
        self._reset_original_values(config_name, orig_value)

    def test_delete(self):
        config_name = 'debug_ms'
        orig_value = self._get_config_by_name(config_name)

        # set a config option
        expected_result = [{'section': 'mon', 'value': '0/3'}]
        self._post('/api/cluster_conf', {
            'name': config_name,
            'value': expected_result
        })
        self.assertStatus(201)
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            expected_result,
            timeout=30,
            period=1)

        # delete it and check if it's deleted
        self._delete('/api/cluster_conf/{}?section={}'.format(config_name, 'mon'))
        self.assertStatus(204)
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            None,
            timeout=30,
            period=1)

        # reset original value
        self._clear_all_values_for_config_option(config_name)
        self._reset_original_values(config_name, orig_value)

    def test_create_cant_update_at_runtime(self):
        config_name = 'public_bind_addr'  # not updatable
        config_value = [{'section': 'global', 'value': 'true'}]
        orig_value = self._get_config_by_name(config_name)

        # try to set config option and check if it fails
        self._post('/api/cluster_conf', {
            'name': config_name,
            'value': config_value
        })
        self.assertStatus(400)
        self.assertError(code='config_option_not_updatable_at_runtime',
                         component='cluster_configuration',
                         detail='Config option {} is/are not updatable at runtime'.format(
                             config_name))

        # check if config option value is still the original one
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            orig_value,
            timeout=30,
            period=1)

    def test_create_two_values(self):
        config_name = 'debug_ms'
        orig_value = self._get_config_by_name(config_name)

        # remove all existing settings for equal preconditions
        self._clear_all_values_for_config_option(config_name)

        expected_result = [{'section': 'mon', 'value': '0/3'},
                           {'section': 'osd', 'value': '0/5'}]

        self._post('/api/cluster_conf', {
            'name': config_name,
            'value': expected_result
        })
        self.assertStatus(201)
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            expected_result,
            timeout=30,
            period=1)

        # reset original value
        self._clear_all_values_for_config_option(config_name)
        self._reset_original_values(config_name, orig_value)

    def test_create_can_handle_none_values(self):
        config_name = 'debug_ms'
        orig_value = self._get_config_by_name(config_name)

        # remove all existing settings for equal preconditions
        self._clear_all_values_for_config_option(config_name)

        self._post('/api/cluster_conf', {
            'name': config_name,
            'value': [{'section': 'mon', 'value': '0/3'},
                      {'section': 'osd', 'value': None}]
        })
        self.assertStatus(201)

        expected_result = [{'section': 'mon', 'value': '0/3'}]
        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            expected_result,
            timeout=30,
            period=1)

        # reset original value
        self._clear_all_values_for_config_option(config_name)
        self._reset_original_values(config_name, orig_value)

    def test_create_can_handle_boolean_values(self):
        config_name = 'mon_allow_pool_delete'
        orig_value = self._get_config_by_name(config_name)

        # remove all existing settings for equal preconditions
        self._clear_all_values_for_config_option(config_name)

        expected_result = [{'section': 'mon', 'value': 'true'}]

        self._post('/api/cluster_conf', {
            'name': config_name,
            'value': [{'section': 'mon', 'value': True}]})
        self.assertStatus(201)

        self.wait_until_equal(
            lambda: self._get_config_by_name(config_name),
            expected_result,
            timeout=30,
            period=1)

        # reset original value
        self._clear_all_values_for_config_option(config_name)
        self._reset_original_values(config_name, orig_value)

    def test_bulk_set(self):
        expected_result = {
            'osd_max_backfills': {'section': 'osd', 'value': '1'},
            'osd_recovery_max_active': {'section': 'osd', 'value': '3'},
            'osd_recovery_max_single_start': {'section': 'osd', 'value': '1'},
            'osd_recovery_sleep': {'section': 'osd', 'value': '2.000000'}
        }
        orig_values = dict()

        for config_name in expected_result:
            orig_values[config_name] = self._get_config_by_name(config_name)

            # remove all existing settings for equal preconditions
            self._clear_all_values_for_config_option(config_name)

        self._put('/api/cluster_conf', {'options': expected_result})
        self.assertStatus(200)

        for config_name, value in expected_result.items():
            self.wait_until_equal(
                lambda: self._get_config_by_name(config_name),
                [value],
                timeout=30,
                period=1)

            # reset original value
            self._clear_all_values_for_config_option(config_name)
            self._reset_original_values(config_name, orig_values[config_name])

    def test_bulk_set_cant_update_at_runtime(self):
        config_options = {
            'public_bind_addr': {'section': 'global', 'value': '1.2.3.4:567'},  # not updatable
            'public_network': {'section': 'global', 'value': '10.0.0.0/8'}  # not updatable
        }
        orig_values = dict()

        for config_name in config_options:
            orig_values[config_name] = self._get_config_by_name(config_name)

        # try to set config options and see if it fails
        self._put('/api/cluster_conf', {'options': config_options})
        self.assertStatus(400)
        self.assertError(code='config_option_not_updatable_at_runtime',
                         component='cluster_configuration',
                         detail='Config option {} is/are not updatable at runtime'.format(
                             ', '.join(config_options.keys())))

        # check if config option values are still the original ones
        for config_name, value in orig_values.items():
            self.wait_until_equal(
                lambda: self._get_config_by_name(config_name),
                value,
                timeout=30,
                period=1)

    def test_bulk_set_cant_update_at_runtime_partial(self):
        config_options = {
            'public_bind_addr': {'section': 'global', 'value': 'true'},  # not updatable
            'log_to_stderr': {'section': 'global', 'value': 'true'}  # updatable
        }
        orig_values = dict()

        for config_name in config_options:
            orig_values[config_name] = self._get_config_by_name(config_name)

        # try to set config options and see if it fails
        self._put('/api/cluster_conf', {'options': config_options})
        self.assertStatus(400)
        self.assertError(code='config_option_not_updatable_at_runtime',
                         component='cluster_configuration',
                         detail='Config option {} is/are not updatable at runtime'.format(
                             'public_bind_addr'))

        # check if config option values are still the original ones
        for config_name, value in orig_values.items():
            self.wait_until_equal(
                lambda: self._get_config_by_name(config_name),
                value,
                timeout=30,
                period=1)

    def test_check_existence(self):
        """
        This test case is intended to check the existence of all hard coded config options used by
        the dashboard.
        If you include further hard coded options in the dashboard, feel free to add them to the
        list.
        """
        hard_coded_options = [
            'osd_max_backfills',  # osd-recv-speed
            'osd_recovery_max_active',  # osd-recv-speed
            'osd_recovery_max_single_start',  # osd-recv-speed
            'osd_recovery_sleep',  # osd-recv-speed
            'osd_scrub_during_recovery',  # osd-pg-scrub
            'osd_scrub_begin_hour',  # osd-pg-scrub
            'osd_scrub_end_hour',  # osd-pg-scrub
            'osd_scrub_begin_week_day',  # osd-pg-scrub
            'osd_scrub_end_week_day',  # osd-pg-scrub
            'osd_scrub_min_interval',  # osd-pg-scrub
            'osd_scrub_max_interval',  # osd-pg-scrub
            'osd_deep_scrub_interval',  # osd-pg-scrub
            'osd_scrub_auto_repair',  # osd-pg-scrub
            'osd_max_scrubs',  # osd-pg-scrub
            'osd_scrub_priority',  # osd-pg-scrub
            'osd_scrub_sleep',  # osd-pg-scrub
            'osd_scrub_auto_repair_num_errors',  # osd-pg-scrub
            'osd_debug_deep_scrub_sleep',  # osd-pg-scrub
            'osd_deep_scrub_keys',  # osd-pg-scrub
            'osd_deep_scrub_large_omap_object_key_threshold',  # osd-pg-scrub
            'osd_deep_scrub_large_omap_object_value_sum_threshold',  # osd-pg-scrub
            'osd_deep_scrub_randomize_ratio',  # osd-pg-scrub
            'osd_deep_scrub_stride',  # osd-pg-scrub
            'osd_deep_scrub_update_digest_min_age',  # osd-pg-scrub
            'osd_requested_scrub_priority',  # osd-pg-scrub
            'osd_scrub_backoff_ratio',  # osd-pg-scrub
            'osd_scrub_chunk_max',  # osd-pg-scrub
            'osd_scrub_chunk_min',  # osd-pg-scrub
            'osd_scrub_cost',  # osd-pg-scrub
            'osd_scrub_interval_randomize_ratio',  # osd-pg-scrub
            'osd_scrub_invalid_stats',  # osd-pg-scrub
            'osd_scrub_load_threshold',  # osd-pg-scrub
            'osd_scrub_max_preemptions',  # osd-pg-scrub
            'mon_allow_pool_delete'  # pool-list
        ]

        for config_option in hard_coded_options:
            self._get('/api/cluster_conf/{}'.format(config_option))
            self.assertStatus(200)

    def _validate_single(self, data):
        self.assertIn('name', data)
        self.assertIn('daemon_default', data)
        self.assertIn('long_desc', data)
        self.assertIn('level', data)
        self.assertIn('default', data)
        self.assertIn('see_also', data)
        self.assertIn('tags', data)
        self.assertIn('min', data)
        self.assertIn('max', data)
        self.assertIn('services', data)
        self.assertIn('type', data)
        self.assertIn('desc', data)
        self.assertIn(data['type'], ['str', 'bool', 'float', 'int', 'size', 'uint', 'addr',
                                     'addrvec', 'uuid', 'secs', 'millisecs'])

        if 'value' in data:
            self.assertIn('source', data)
            self.assertIsInstance(data['value'], list)

            for entry in data['value']:
                self.assertIsInstance(entry, dict)
                self.assertIn('section', entry)
                self.assertIn('value', entry)

    def _get_config_by_name(self, conf_name):
        data = self._get('/api/cluster_conf/{}'.format(conf_name))
        if 'value' in data:
            return data['value']
        return None

    def _clear_all_values_for_config_option(self, config_name):
        values = self._get_config_by_name(config_name)
        if values:
            for value in values:
                self._ceph_cmd(['config', 'rm', value['section'], config_name])

    def _reset_original_values(self, config_name, orig_values):
        if orig_values:
            for value in orig_values:
                self._ceph_cmd(['config', 'set', value['section'], config_name, value['value']])



"""
Test for Ceph clusters with multiple FSs.
"""
import logging

from tasks.cephfs.cephfs_test_case import CephFSTestCase
from tasks.cephfs.caps_helper import (MonCapTester, MdsCapTester,
                                      gen_mon_cap_str, gen_osd_cap_str,
                                      gen_mds_cap_str)

from teuthology.exceptions import CommandFailedError


log = logging.getLogger(__name__)


class TestMultiFS(CephFSTestCase):
    client_id = 'testuser'
    client_name = 'client.' + client_id
    # one dedicated for each FS
    MDSS_REQUIRED = 2
    CLIENTS_REQUIRED = 2

    def setUp(self):
        super(TestMultiFS, self).setUp()

        # we might have it - the client - if the same cluster was used for a
        # different vstart_runner.py run.
        self.run_ceph_cmd(f'auth rm {self.client_name}')

        self.fs1 = self.fs
        self.fs2 = self.mds_cluster.newfs(name='cephfs2', create=True)

        # we'll reassign caps to client.1 so that it can operate with cephfs2
        self.run_ceph_cmd(f'auth caps client.{self.mount_b.client_id} mon '
                          f'"allow r" osd "allow rw '
                          f'pool={self.fs2.get_data_pool_name()}" mds allow')
        self.mount_b.remount(cephfs_name=self.fs2.name)


class TestMONCaps(TestMultiFS):

    def test_moncap_with_one_fs_names(self):
        self.captester = MonCapTester()
        moncap = gen_mon_cap_str((('r', self.fs1.name),))
        self.create_client(self.client_id, moncap)

        self.captester.run_mon_cap_tests(self.fs1, self.client_id)

    def test_moncap_with_multiple_fs_names(self):
        self.captester = MonCapTester()
        moncap = gen_mon_cap_str((('r', self.fs1.name), ('r', self.fs2.name)))
        self.create_client(self.client_id, moncap)

        self.captester.run_mon_cap_tests(self.fs1, self.client_id)

    def test_moncap_with_blanket_allow(self):
        self.captester = MonCapTester()
        moncap = 'allow r'
        self.create_client(self.client_id, moncap)

        self.captester.run_mon_cap_tests(self.fs1, self.client_id)


#TODO: add tests for capsecs 'p' and 's'.
class TestMDSCaps(TestMultiFS):
    """
    0. Have 2 FSs on Ceph cluster.
    1. Create new files on both FSs.
    2. Create a new client that has authorization for both FSs.
    3. Remount the current mounts with this new client.
    4. Test read and write on both FSs.
    """
    def test_rw_with_fsname_and_no_path_in_cap(self):
        PERM = 'rw'
        self.captesters = (MdsCapTester(self.mount_a),
                           MdsCapTester(self.mount_b))

        moncap, osdcap, mdscap = self._gen_caps(PERM, both_fsnames=True)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring)

        self.captesters[0].run_mds_cap_tests(PERM)
        self.captesters[1].run_mds_cap_tests(PERM)

    def test_r_with_fsname_and_no_path_in_cap(self):
        PERM = 'r'
        self.captesters = (MdsCapTester(self.mount_a),
                           MdsCapTester(self.mount_b))

        moncap, osdcap, mdscap = self._gen_caps(PERM, both_fsnames=True)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring)

        self.captesters[0].run_mds_cap_tests(PERM)
        self.captesters[1].run_mds_cap_tests(PERM)

    def test_rw_with_fsname_and_path_in_cap(self):
        PERM, CEPHFS_MNTPT = 'rw', 'dir1'
        self.mount_a.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.mount_b.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.captesters = (MdsCapTester(self.mount_a, CEPHFS_MNTPT),
                           MdsCapTester(self.mount_b, CEPHFS_MNTPT))

        moncap, osdcap, mdscap = self._gen_caps(PERM, True, CEPHFS_MNTPT)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring, CEPHFS_MNTPT)

        self.captesters[0].run_mds_cap_tests(PERM, CEPHFS_MNTPT)
        self.captesters[1].run_mds_cap_tests(PERM, CEPHFS_MNTPT)

    def test_r_with_fsname_and_path_in_cap(self):
        PERM, CEPHFS_MNTPT = 'r', 'dir1'
        self.mount_a.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.mount_b.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.captesters = (MdsCapTester(self.mount_a, CEPHFS_MNTPT),
                           MdsCapTester(self.mount_b, CEPHFS_MNTPT))

        moncap, osdcap, mdscap = self._gen_caps(PERM, True, CEPHFS_MNTPT)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring, CEPHFS_MNTPT)

        self.captesters[0].run_mds_cap_tests(PERM, CEPHFS_MNTPT)
        self.captesters[1].run_mds_cap_tests(PERM, CEPHFS_MNTPT)

    # XXX: this tests the backward compatibility; "allow rw path=<dir1>" is
    # treated as "allow rw fsname=* path=<dir1>"
    def test_rw_with_no_fsname_and_path_in_cap(self):
        PERM, CEPHFS_MNTPT = 'rw', 'dir1'
        self.mount_a.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.mount_b.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.captesters = (MdsCapTester(self.mount_a, CEPHFS_MNTPT),
                           MdsCapTester(self.mount_b, CEPHFS_MNTPT))

        moncap, osdcap, mdscap = self._gen_caps(PERM, False, CEPHFS_MNTPT)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring, CEPHFS_MNTPT)

        self.captesters[0].run_mds_cap_tests(PERM, CEPHFS_MNTPT)
        self.captesters[1].run_mds_cap_tests(PERM, CEPHFS_MNTPT)

    # XXX: this tests the backward compatibility; "allow r path=<dir1>" is
    # treated as "allow r fsname=* path=<dir1>"
    def test_r_with_no_fsname_and_path_in_cap(self):
        PERM, CEPHFS_MNTPT = 'r', 'dir1'
        self.mount_a.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.mount_b.run_shell(f'mkdir {CEPHFS_MNTPT}')
        self.captesters = (MdsCapTester(self.mount_a, CEPHFS_MNTPT),
                           MdsCapTester(self.mount_b, CEPHFS_MNTPT))

        moncap, osdcap, mdscap = self._gen_caps(PERM, False, CEPHFS_MNTPT)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring, CEPHFS_MNTPT)

        self.captesters[0].run_mds_cap_tests(PERM, CEPHFS_MNTPT)
        self.captesters[1].run_mds_cap_tests(PERM, CEPHFS_MNTPT)

    def test_rw_with_no_fsname_and_no_path(self):
        PERM = 'rw'
        self.captesters = (MdsCapTester(self.mount_a),
                           MdsCapTester(self.mount_b))

        moncap, osdcap, mdscap = self._gen_caps(PERM)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring)

        self.captesters[0].run_mds_cap_tests(PERM)
        self.captesters[1].run_mds_cap_tests(PERM)

    def test_r_with_no_fsname_and_no_path(self):
        PERM = 'r'
        self.captesters = (MdsCapTester(self.mount_a),
                           MdsCapTester(self.mount_b))

        moncap, osdcap, mdscap = self._gen_caps(PERM)
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        self.remount_with_new_client(keyring)

        self.captesters[0].run_mds_cap_tests(PERM)
        self.captesters[1].run_mds_cap_tests(PERM)

    def tearDown(self):
        self.mount_a.umount_wait()
        self.mount_b.umount_wait()

        super(type(self), self).tearDown()

    def _gen_caps(self, perm, both_fsnames=False, cephfs_mntpt='/'):
        moncap = 'allow r'
        osdcap = gen_osd_cap_str(((perm, self.fs1.name),
                                  (perm, self.fs2.name)))

        if both_fsnames:
            mdscap = gen_mds_cap_str(((perm, self.fs1.name, cephfs_mntpt),
                                      (perm, self.fs2.name, cephfs_mntpt)))
        else:
            mdscap = gen_mds_cap_str(((perm, None, cephfs_mntpt),
                                      (perm, None, cephfs_mntpt)))

        return moncap, osdcap, mdscap

    def remount_with_new_client(self, keyring, cephfs_mntpt='/'):
        log.info(f'keyring generated for testing is -\n{keyring}')

        if isinstance(cephfs_mntpt, str) and cephfs_mntpt != '/' :
            cephfs_mntpt = '/' + cephfs_mntpt

        keyring_path = self.mount_a.client_remote.mktemp(data=keyring)
        self.mount_a.remount(client_id=self.client_id,
                             client_keyring_path=keyring_path,
                             client_remote=self.mount_a.client_remote,
                             cephfs_name=self.fs1.name,
                             cephfs_mntpt=cephfs_mntpt,
                             hostfs_mntpt=self.mount_a.hostfs_mntpt,
                             wait=True)

        keyring_path = self.mount_b.client_remote.mktemp(data=keyring)
        self.mount_b.remount(client_id=self.client_id,
                             client_keyring_path=keyring_path,
                             client_remote=self.mount_b.client_remote,
                             cephfs_name=self.fs2.name,
                             cephfs_mntpt=cephfs_mntpt,
                             hostfs_mntpt=self.mount_b.hostfs_mntpt,
                             wait=True)


class TestClientsWithoutAuth(TestMultiFS):
    # c.f., src/mount/mtab.c: EX_FAIL
    RETVAL_KCLIENT = 32
    # c.f., src/ceph_fuse.cc: (cpp EXIT_FAILURE). Normally the check for this
    # case should be anything-except-0, but EXIT_FAILURE is 1 in most systems.
    RETVAL_USER_SPACE_CLIENT = 1

    def setUp(self):
        super(TestClientsWithoutAuth, self).setUp()
        self.retval = self.RETVAL_KCLIENT if 'kernel' in str(type(self.mount_a)).lower() \
            else self.RETVAL_USER_SPACE_CLIENT

    def test_mount_all_caps_absent(self):
        # setup part...
        keyring = self.fs1.authorize(self.client_id, ('/', 'rw'))
        keyring_path = self.mount_a.client_remote.mktemp(data=keyring)

        # mount the FS for which client has no auth...
        try:
            self.mount_a.remount(client_id=self.client_id,
                                 client_keyring_path=keyring_path,
                                 cephfs_name=self.fs2.name,
                                 check_status=False)
        except CommandFailedError as e:
            self.assertEqual(e.exitstatus, self.retval)

    def test_mount_mon_and_osd_caps_present_mds_caps_absent(self):
        # setup part...
        moncap = gen_mon_cap_str((('rw', self.fs1.name),
                                  ('rw', self.fs2.name)))
        mdscap = gen_mds_cap_str((('rw', self.fs1.name),))
        osdcap = gen_osd_cap_str((('rw', self.fs1.name,),
                                  ('rw', self.fs2.name)))
        keyring = self.create_client(self.client_id, moncap, osdcap, mdscap)
        keyring_path = self.mount_a.client_remote.mktemp(data=keyring)

        # mount the FS for which client has no auth...
        try:
            self.mount_a.remount(client_id=self.client_id,
                                 client_keyring_path=keyring_path,
                                 cephfs_name=self.fs2.name,
                                 check_status=False)
        except CommandFailedError as e:
            self.assertEqual(e.exitstatus, self.retval)



"""
Deploy and configure Tempest for Teuthology
"""
import configparser
import contextlib
import logging

from teuthology import misc as teuthology
from teuthology import contextutil
from teuthology.exceptions import ConfigError
from teuthology.orchestra import run

log = logging.getLogger(__name__)


def get_tempest_dir(ctx):
    return '{tdir}/tempest'.format(tdir=teuthology.get_testdir(ctx))

def run_in_tempest_dir(ctx, client, cmdargs, **kwargs):
    ctx.cluster.only(client).run(
        args=[ 'cd', get_tempest_dir(ctx), run.Raw('&&'), ] + cmdargs,
        **kwargs
    )

def run_in_tempest_rgw_dir(ctx, client, cmdargs, **kwargs):
    ctx.cluster.only(client).run(
        args=[ 'cd', get_tempest_dir(ctx) + '/rgw', run.Raw('&&'), ] + cmdargs,
        **kwargs
    )

def run_in_tempest_venv(ctx, client, cmdargs, **kwargs):
    run_in_tempest_dir(ctx, client,
                        [   'source',
                            '.tox/venv/bin/activate',
                            run.Raw('&&')
                        ] + cmdargs, **kwargs)

@contextlib.contextmanager
def download(ctx, config):
    """
    Download the Tempest from github.
    Remove downloaded file upon exit.

    The context passed in should be identical to the context
    passed in to the main task.
    """
    assert isinstance(config, dict)
    log.info('Downloading Tempest...')
    for (client, cconf) in config.items():
        ctx.cluster.only(client).run(
            args=[
                'git', 'clone',
                '-b', cconf.get('force-branch', 'master'),
                'https://github.com/openstack/tempest.git',
                get_tempest_dir(ctx)
            ],
        )

        sha1 = cconf.get('sha1')
        if sha1 is not None:
            run_in_tempest_dir(ctx, client, [ 'git', 'reset', '--hard', sha1 ])
    try:
        yield
    finally:
        log.info('Removing Tempest...')
        for client in config:
            ctx.cluster.only(client).run(
                args=[ 'rm', '-rf', get_tempest_dir(ctx) ],
            )

def get_toxvenv_dir(ctx):
    return ctx.tox.venv_path

@contextlib.contextmanager
def setup_venv(ctx, config):
    """
    Setup the virtualenv for Tempest using tox.
    """
    assert isinstance(config, dict)
    log.info('Setting up virtualenv for Tempest')
    for (client, _) in config.items():
        run_in_tempest_dir(ctx, client,
            [   '{tvdir}/bin/tox'.format(tvdir=get_toxvenv_dir(ctx)),
                '-e', 'venv', '--notest'
            ])
    yield

def setup_logging(ctx, cpar):
    cpar.set('DEFAULT', 'log_dir', teuthology.get_archive_dir(ctx))
    cpar.set('DEFAULT', 'log_file', 'tempest.log')

def to_config(config, params, section, cpar):
    for (k, v) in config[section].items():
        if isinstance(v, str):
            v = v.format(**params)
        elif isinstance(v, bool):
            v = 'true' if v else 'false'
        else:
            v = str(v)
        cpar.set(section, k, v)

@contextlib.contextmanager
def configure_instance(ctx, config):
    assert isinstance(config, dict)
    log.info('Configuring Tempest')

    for (client, cconfig) in config.items():
        run_in_tempest_venv(ctx, client,
            [
                'tempest',
                'init',
                '--workspace-path',
                get_tempest_dir(ctx) + '/workspace.yaml',
                'rgw'
            ])

        # prepare the config file
        tetcdir = '{tdir}/rgw/etc'.format(tdir=get_tempest_dir(ctx))
        (remote,) = ctx.cluster.only(client).remotes.keys()
        local_conf = remote.get_file(tetcdir + '/tempest.conf.sample')

        # fill the params dictionary which allows to use templatized configs
        keystone_role = cconfig.get('use-keystone-role', None)
        if keystone_role is None \
            or keystone_role not in ctx.keystone.public_endpoints:
            raise ConfigError('the use-keystone-role is misconfigured')
        public_host, public_port = ctx.keystone.public_endpoints[keystone_role]
        params = {
            'keystone_public_host': public_host,
            'keystone_public_port': str(public_port),
        }

        cpar = configparser.ConfigParser()
        cpar.read(local_conf)
        setup_logging(ctx, cpar)
        to_config(cconfig, params, 'auth', cpar)
        to_config(cconfig, params, 'identity', cpar)
        to_config(cconfig, params, 'object-storage', cpar)
        to_config(cconfig, params, 'object-storage-feature-enabled', cpar)
        cpar.write(open(local_conf, 'w+'))

        remote.put_file(local_conf, tetcdir + '/tempest.conf')
    yield

@contextlib.contextmanager
def run_tempest(ctx, config):
    assert isinstance(config, dict)
    log.info('Configuring Tempest')

    for (client, cconf) in config.items():
        blocklist = cconf.get('blocklist', [])
        assert isinstance(blocklist, list)
        run_in_tempest_venv(ctx, client,
            [
                'tempest',
                'run',
                '--workspace-path',
                get_tempest_dir(ctx) + '/workspace.yaml',
                '--workspace',
                'rgw',
                '--regex', '^tempest.api.object_storage',
                '--black-regex', '|'.join(blocklist)
            ])
    try:
        yield
    finally:
        pass


@contextlib.contextmanager
def task(ctx, config):
    """
    Deploy and run Tempest's object storage campaign

    Example of configuration:

      overrides:
        ceph:
          conf:
            client:
              rgw keystone api version: 3
              rgw keystone accepted roles: admin,member
              rgw keystone implicit tenants: true
              rgw keystone accepted admin roles: admin
              rgw swift enforce content length: true
              rgw swift account in url: true
              rgw swift versioning enabled: true
              rgw keystone admin domain: Default
              rgw keystone admin user: admin
              rgw keystone admin password: ADMIN
              rgw keystone admin project: admin
      tasks:
      # typically, the task should be preceded with install, ceph, tox,
      # keystone and rgw. Tox and Keystone are specific requirements
      # of tempest.py.
      - rgw:
          # it's important to match the prefix with the endpoint's URL
          # in Keystone. Additionally, if we want to test /info and its
          # accompanying stuff, the whole Swift API must be put in root
          # of the whole URL  hierarchy (read: frontend_prefix == /swift).
          frontend_prefix: /swift
          client.0:
            use-keystone-role: client.0
      - tempest:
          client.0:
            force-branch: master
            use-keystone-role: client.0
            auth:
              admin_username: admin
              admin_project_name: admin
              admin_password: ADMIN
              admin_domain_name: Default
            identity:
              uri: http://{keystone_public_host}:{keystone_public_port}/v2.0/
              uri_v3: http://{keystone_public_host}:{keystone_public_port}/v3/
              admin_role: admin
            object-storage:
              reseller_admin_role: admin
            object-storage-feature-enabled:
              container_sync: false
              discoverability: false
            blocklist:
              # please strip half of these items after merging PRs #15369
              # and #12704
              - .*test_list_containers_reverse_order.*
              - .*test_list_container_contents_with_end_marker.*
              - .*test_delete_non_empty_container.*
              - .*test_container_synchronization.*
              - .*test_get_object_after_expiration_time.*
              - .*test_create_object_with_transfer_encoding.*
    """
    assert config is None or isinstance(config, list) \
        or isinstance(config, dict), \
        'task tempest only supports a list or dictionary for configuration'

    if not ctx.tox:
        raise ConfigError('tempest must run after the tox task')
    if not ctx.keystone:
        raise ConfigError('tempest must run after the keystone task')

    all_clients = ['client.{id}'.format(id=id_)
                   for id_ in teuthology.all_roles_of_type(ctx.cluster, 'client')]
    if config is None:
        config = all_clients
    if isinstance(config, list):
        config = dict.fromkeys(config)

    overrides = ctx.config.get('overrides', {})
    # merge each client section, not the top level.
    for client in config.keys():
        if not config[client]:
            config[client] = {}
        teuthology.deep_merge(config[client], overrides.get('keystone', {}))

    log.debug('Tempest config is %s', config)

    with contextutil.nested(
        lambda: download(ctx=ctx, config=config),
        lambda: setup_venv(ctx=ctx, config=config),
        lambda: configure_instance(ctx=ctx, config=config),
        lambda: run_tempest(ctx=ctx, config=config),
        ):
        yield



#!/usr/bin/env python3

#
# Copyright(c) 2012-2018 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause-Clear
#

import tests_config
import os
import sys
import subprocess


def run_command(args):
    result = subprocess.run(" ".join(args), shell=True,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result.stdout = result.stdout.decode("ASCII", errors='ignore')
    result.stderr = result.stderr.decode("ASCII", errors='ignore')
    return result


def rmv_cmd(trgt):
    """Remove target with force"""
    result = run_command(["rm", "-rf", trgt])
    if result.returncode != 0:
        raise Exception("Removing {} before testing failed!".
                        format(os.path.dirname(os.path.realpath(__file__))
                               + trgt))


def cleanup():
    """Delete files created by unit tests"""
    script_path = os.path.dirname(os.path.realpath(__file__))
    test_dir = os.path.join(script_path, tests_config.MAIN_DIRECTORY_OF_UNIT_TESTS)
    result = run_command(["cd", test_dir])
    if result.returncode != 0:
        raise Exception("Cleanup before testing failed!")

    # r=root, d=directories, f = files
    for r, d, f in os.walk(test_dir):
        for file in f:
            if '_generated_wrap' in file:
                rmv_cmd(file)

    rmv_cmd("preprocessed_sources_repository")
    rmv_cmd("sources_to_test_repository")
    rmv_cmd("build")

    result = run_command(["cd", script_path])
    if result.returncode != 0:
        raise Exception("Cleanup before testing failed!")


cleanup()

script_path = os.path.dirname(os.path.realpath(__file__))

main_UT_dir = os.path.join(script_path, tests_config.MAIN_DIRECTORY_OF_UNIT_TESTS)

main_env_dir = os.path.join(script_path, tests_config.MAIN_DIRECTORY_OF_ENV_FILES)

main_tested_dir = os.path.join(script_path, tests_config.MAIN_DIRECTORY_OF_TESTED_PROJECT)

if not os.path.isdir(os.path.join(main_UT_dir, "ocf_env", "ocf")):
    try:
        os.makedirs(os.path.join(main_UT_dir, "ocf_env", "ocf"))
    except Exception:
        raise Exception("Cannot create ocf_env/ocf directory!")

result = run_command(["ln", "-fs",
                      os.path.join(main_env_dir, "*"),
                      os.path.join(main_UT_dir, "ocf_env")])
if result.returncode != 0:
    raise Exception("Preparing env sources for testing failed!")

result = run_command(["cp", "-r",
                      os.path.join(main_tested_dir, "inc", "*"),
                      os.path.join(main_UT_dir, "ocf_env", "ocf")])
if result.returncode != 0:
    raise Exception("Preparing sources for testing failed!")

result = run_command([os.path.join(script_path, "prepare_sources_for_testing.py")])
if result.returncode != 0:
    raise Exception("Preparing sources for testing failed!")

build_dir = os.path.join(main_UT_dir, "build")
logs_dir = os.path.join(main_UT_dir, "logs")

try:
    if not os.path.isdir(build_dir):
        os.makedirs(build_dir)
    if not os.path.isdir(logs_dir):
        os.makedirs(logs_dir)
except Exception:
    raise Exception("Cannot create logs directory!")

os.chdir(build_dir)

cmake_result = run_command(["cmake", ".."])

print(cmake_result.stdout)
with open(os.path.join(logs_dir, "cmake.output"), "w") as f:
    f.write(cmake_result.stdout)
    f.write(cmake_result.stderr)

if cmake_result.returncode != 0:
    with open(os.path.join(logs_dir, "tests.output"), "w") as f:
        f.write("Cmake step failed! More details in cmake.output.")
    sys.exit(1)

make_result = run_command(["make", "-j"])

print(make_result.stdout)
with open(os.path.join(logs_dir, "make.output"), "w") as f:
    f.write(make_result.stdout)
    f.write(make_result.stderr)

if make_result.returncode != 0:
    with open(os.path.join(logs_dir, "tests.output"), "w") as f:
        f.write("Make step failed! More details in make.output.")
    sys.exit(1)

test_result = run_command(["make", "test"])

print(test_result.stdout)
with open(os.path.join(logs_dir, "tests.output"), "w") as f:
    f.write(test_result.stdout)



from rpc.client import JSONRPCException
from .ui_node import UINode


class UINVMf(UINode):
    def __init__(self, parent):
        UINode.__init__(self, "nvmf", parent)
        self.refresh()

    def refresh(self):
        self._children = set([])
        UINVMfSubsystems(self)
        UINVMfTransports(self)


class UINVMfTransports(UINode):
    def __init__(self, parent):
        UINode.__init__(self, "transport", parent)
        self.refresh()

    def refresh(self):
        self._children = set([])
        for transport in self.get_root().nvmf_get_transports():
            UINVMfTransport(transport, self)

    def ui_command_create(self, trtype, max_queue_depth=None, max_io_qpairs_per_ctrlr=None,
                          in_capsule_data_size=None, max_io_size=None, io_unit_size=None, max_aq_depth=None):
        """Create a transport with given parameters

        Arguments:
            trtype - Example: 'RDMA'.
            max_queue_depth - Optional parameter. Integer, max value 65535.
            max_io_qpairs_per_ctrlr - Optional parameter. 16 bit Integer, max value 65535.
            in_capsule_data_size - Optional parameter. 32 bit Integer, max value 4294967295
            max_io_size - Optional parameter. 32 bit integer, max value 4294967295
            io_unit_size - Optional parameter. 32 bit integer, max value 4294967295
            max_aq_depth - Optional parameter. 32 bit integer, max value 4294967295
        """
        max_queue_depth = self.ui_eval_param(max_queue_depth, "number", None)
        max_io_qpairs_per_ctrlr = self.ui_eval_param(max_io_qpairs_per_ctrlr, "number", None)
        in_capsule_data_size = self.ui_eval_param(in_capsule_data_size, "number", None)
        max_io_size = self.ui_eval_param(max_io_size, "number", None)
        io_unit_size = self.ui_eval_param(io_unit_size, "number", None)
        max_aq_depth = self.ui_eval_param(max_aq_depth, "number", None)

        self.get_root().create_nvmf_transport(trtype=trtype,
                                              max_queue_depth=max_queue_depth,
                                              max_io_qpairs_per_ctrlr=max_io_qpairs_per_ctrlr,
                                              in_capsule_data_size=in_capsule_data_size,
                                              max_io_size=max_io_size,
                                              io_unit_size=io_unit_size,
                                              max_aq_depth=max_aq_depth)

    def summary(self):
        return "Transports: %s" % len(self.children), None


class UINVMfTransport(UINode):
    def __init__(self, transport, parent):
        UINode.__init__(self, transport.trtype, parent)
        self.transport = transport


class UINVMfSubsystems(UINode):
    def __init__(self, parent):
        UINode.__init__(self, "subsystem", parent)
        self.refresh()

    def refresh(self):
        self._children = set([])
        for subsystem in self.get_root().nvmf_get_subsystems():
            UINVMfSubsystem(subsystem, self)

    def delete(self, subsystem_nqn):
        self.get_root().nvmf_delete_subsystem(nqn=subsystem_nqn)

    def ui_command_create(self, nqn, serial_number=None,
                          max_namespaces=None, allow_any_host="false"):
        """Create subsystem with given parameteres.

        Arguments:
            nqn - Target nqn(ASCII).
            serial_number - Example: 'SPDK00000000000001'.
            max_namespaces - Optional parameter. Maximum number of namespaces allowed to added during
                             active connection
            allow_any_host - Optional parameter. Allow any host to connect (don't enforce host NQN
                             whitelist)
        """
        allow_any_host = self.ui_eval_param(allow_any_host, "bool", False)
        max_namespaces = self.ui_eval_param(max_namespaces, "number", 0)
        self.get_root().create_nvmf_subsystem(nqn=nqn, serial_number=serial_number,
                                              allow_any_host=allow_any_host,
                                              max_namespaces=max_namespaces)

    def ui_command_delete(self, subsystem_nqn):
        """Delete subsystem with given nqn.

        Arguments:
            nqn_subsystem - Name of susbsytem to delete
        """
        self.delete(subsystem_nqn)

    def ui_command_delete_all(self):
        """Delete all subsystems"""
        rpc_messages = ""
        for child in self._children:
            try:
                self.delete(child.subsystem.nqn)
            except JSONRPCException as e:
                rpc_messages += e.message
        if rpc_messages:
            raise JSONRPCException(rpc_messages)

    def summary(self):
        return "Subsystems: %s" % len(self.children), None


class UINVMfSubsystem(UINode):
    def __init__(self, subsystem, parent):
        UINode.__init__(self, subsystem.nqn, parent)
        self.subsystem = subsystem
        self.refresh()

    def refresh(self):
        self._children = set([])
        UINVMfSubsystemListeners(self.subsystem.listen_addresses, self)
        UINVMfSubsystemHosts(self.subsystem.hosts, self)
        if hasattr(self.subsystem, 'namespaces'):
            UINVMfSubsystemNamespaces(self.subsystem.namespaces, self)

    def refresh_node(self):
        for subsystem in self.get_root().nvmf_get_subsystems():
            if subsystem.nqn == self.subsystem.nqn:
                self.subsystem = subsystem
        self.refresh()

    def ui_command_show_details(self):
        self.shell.log.info(json.dumps(vars(self.lvs), indent=2))

    def ui_command_allow_any_host(self, disable="false"):
        """Disable or or enable allow_any_host flag.

        Arguments:
            disable - Optional parameter. If false then enable, if true disable
        """
        disable = self.ui_eval_param(disable, "bool", None)
        self.get_root().nvmf_subsystem_allow_any_host(
            nqn=self.subsystem.nqn, disable=disable)

    def summary(self):
        sn = None
        if hasattr(self.subsystem, 'serial_number'):
            sn = "sn=%s" % self.subsystem.serial_number
        st = None
        if hasattr(self.subsystem, 'subtype'):
            st = "st=%s" % self.subsystem.subtype
        allow_any_host = None
        if self.subsystem.allow_any_host:
            allow_any_host = "Allow any host"
        info = ", ".join(filter(None, [sn, st, allow_any_host]))
        return info, None


class UINVMfSubsystemListeners(UINode):
    def __init__(self, listen_addresses, parent):
        UINode.__init__(self, "listen_addresses", parent)
        self.listen_addresses = listen_addresses
        self.refresh()

    def refresh(self):
        self._children = set([])
        for address in self.listen_addresses:
            UINVMfSubsystemListener(address, self)

    def refresh_node(self):
        for subsystem in self.get_root().nvmf_get_subsystems():
            if subsystem.nqn == self.parent.subsystem.nqn:
                self.listen_addresses = subsystem.listen_addresses
        self.refresh()

    def delete(self, trtype, traddr, trsvcid, adrfam=None):
        self.get_root().nvmf_subsystem_remove_listener(
            nqn=self.parent.subsystem.nqn, trtype=trtype,
            traddr=traddr, trsvcid=trsvcid, adrfam=adrfam)

    def ui_command_create(self, trtype, traddr, trsvcid, adrfam):
        """Create address listener for subsystem.

        Arguments:
            trtype - NVMe-oF transport type: e.g., rdma.
            traddr - NVMe-oF transport address: e.g., an ip address.
            trsvcid - NVMe-oF transport service id: e.g., a port number.
            adrfam - NVMe-oF transport adrfam: e.g., ipv4, ipv6, ib, fc.
        """
        self.get_root().nvmf_subsystem_add_listener(
            nqn=self.parent.subsystem.nqn, trtype=trtype, traddr=traddr,
            trsvcid=trsvcid, adrfam=adrfam)

    def ui_command_delete(self, trtype, traddr, trsvcid, adrfam=None):
        """Remove address listener for subsystem.

        Arguments:
            trtype - Transport type (RDMA)
            traddr - NVMe-oF transport address: e.g., an ip address.
            trsvcid - NVMe-oF transport service id: e.g., a port number.
            adrfam - Optional argument. Address family ("IPv4", "IPv6", "IB" or "FC").
        """
        self.delete(trtype, traddr, trsvcid, adrfam)

    def ui_command_delete_all(self):
        """Remove all address listeners from subsystem."""
        rpc_messages = ""
        for la in self.listen_addresses:
            try:
                self.delete(la['trtype'], la['traddr'], la['trsvcid'], la['adrfam'])
            except JSONRPCException as e:
                rpc_messages += e.message
        if rpc_messages:
            raise JSONRPCException(rpc_messages)

    def summary(self):
        return "Addresses: %s" % len(self.listen_addresses), None


class UINVMfSubsystemListener(UINode):
    def __init__(self, address, parent):
        UINode.__init__(self, "%s:%s" % (address['traddr'], address['trsvcid']),
                        parent)
        self.address = address

    def summary(self):
        return "%s" % self.address['trtype'], True


class UINVMfSubsystemHosts(UINode):
    def __init__(self, hosts, parent):
        UINode.__init__(self, "hosts", parent)
        self.hosts = hosts
        self.refresh()

    def refresh(self):
        self._children = set([])
        for host in self.hosts:
            UINVMfSubsystemHost(host, self)

    def refresh_node(self):
        for subsystem in self.get_root().nvmf_get_subsystems():
            if subsystem.nqn == self.parent.subsystem.nqn:
                self.hosts = subsystem.hosts
        self.refresh()

    def delete(self, host):
        self.get_root().nvmf_subsystem_remove_host(
            nqn=self.parent.subsystem.nqn, host=host)

    def ui_command_create(self, host):
        """Add a host NQN to the whitelist of allowed hosts.

        Args:
            host: Host NQN to add to the list of allowed host NQNs
        """
        self.get_root().nvmf_subsystem_add_host(
            nqn=self.parent.subsystem.nqn, host=host)

    def ui_command_delete(self, host):
        """Delete host from subsystem.

        Arguments:
           host - NQN of host to remove.
        """
        self.delete(host)

    def ui_command_delete_all(self):
        """Delete host from subsystem"""
        rpc_messages = ""
        for host in self.hosts:
            try:
                self.delete(host['nqn'])
            except JSONRPCException as e:
                rpc_messages += e.message
        if rpc_messages:
            raise JSONRPCException(rpc_messages)

    def summary(self):
        return "Hosts: %s" % len(self.hosts), None


class UINVMfSubsystemHost(UINode):
    def __init__(self, host, parent):
        UINode.__init__(self, "%s" % host['nqn'], parent)
        self.host = host


class UINVMfSubsystemNamespaces(UINode):
    def __init__(self, namespaces, parent):
        UINode.__init__(self, "namespaces", parent)
        self.namespaces = namespaces
        self.refresh()

    def refresh(self):
        self._children = set([])
        for namespace in self.namespaces:
            UINVMfSubsystemNamespace(namespace, self)

    def refresh_node(self):
        for subsystem in self.get_root().nvmf_get_subsystems():
            if subsystem.nqn == self.parent.subsystem.nqn:
                self.namespaces = subsystem.namespaces
        self.refresh()

    def delete(self, nsid):
        self.get_root().nvmf_subsystem_remove_ns(
            nqn=self.parent.subsystem.nqn, nsid=nsid)

    def ui_command_create(self, bdev_name, nsid=None,
                          nguid=None, eui64=None, uuid=None):
        """Add a namespace to a subsystem.

        Args:
            bdev_name: Name of bdev to expose as a namespace.
        Optional args:
            nsid: Namespace ID.
            nguid: 16-byte namespace globally unique identifier in hexadecimal.
            eui64: 8-byte namespace EUI-64 in hexadecimal (e.g. "ABCDEF0123456789").
            uuid: Namespace UUID.
        """
        nsid = self.ui_eval_param(nsid, "number", None)
        self.get_root().nvmf_subsystem_add_ns(
            nqn=self.parent.subsystem.nqn, bdev_name=bdev_name,
            nsid=nsid, nguid=nguid, eui64=eui64, uuid=uuid)

    def ui_command_delete(self, nsid):
        """Delete namespace from subsystem.

        Arguments:
            nsid - Id of namespace to remove.
        """
        nsid = self.ui_eval_param(nsid, "number", None)
        self.delete(nsid)

    def ui_command_delete_all(self):
        """Delete all namespaces from subsystem."""
        rpc_messages = ""
        for namespace in self.namespaces:
            try:
                self.delete(namespace['nsid'])
            except JSONRPCException as e:
                rpc_messages += e.message
        if rpc_messages:
            raise JSONRPCException(rpc_messages)

    def summary(self):
        return "Namespaces: %s" % len(self.namespaces), None


class UINVMfSubsystemNamespace(UINode):
    def __init__(self, namespace, parent):
        UINode.__init__(self, namespace['bdev_name'], parent)
        self.namespace = namespace

    def summary(self):
        info = ", ".join([self.namespace['name'], str(self.namespace['nsid'])])
        return info, None
